"""
Constants for csle-common
"""
from typing import Union
import re
import json
from csle_common.dao.emulation_config.config import Config


class GENERAL:
    """
    General constants
    """
    THREAD_PROPERTY = "thread"
    PORT_PROPERTY = "port"
    IP_PROPERTY = "ip"
    EMULATION_PROPERTY = "emulation"
    EXECUTION_ID_PROPERTY = "execution_id"


class GRPC_SERVERS:
    """
    Constants related to grpc servers
    """
    CLUSTER_MANAGER_PORT = 50041
    RYU_MANAGER_PORT = 50042
    TRAFFIC_MANAGER_PORT = 50043
    CLIENT_MANAGER_PORT = 50044
    ELK_MANAGER_PORT = 50045
    DOCKER_STATS_MANAGER_PORT = 50046
    OSSEC_IDS_MANAGER_PORT = 50047
    SNORT_IDS_MANAGER_PORT = 50048
    HOST_MANAGER_PORT = 50049
    KAFKA_MANAGER_DEFAULT_PORT = 50051
    FIVE_G_CORE_MANAGER_DEFAULT_PORT = 50052
    FIVE_G_CU_MANAGER_DEFAULT_PORT = 50053
    FIVE_G_DU_MANAGER_DEFAULT_PORT = 50054
    SERVICE_CONFIG_JSON = json.dumps({
        "methodConfig": [{
            # An empty "name" list means this policy applies to all methods.
            "name": [{}],
            "retryPolicy": {
                "maxAttempts": 5,  # Try a total of 5 times
                "initialBackoff": "0.1s",  # Start with a 100ms delay
                "maxBackoff": "10s",  # Maximum delay is 10 seconds
                "backoffMultiplier": 2,  # Double the delay each time
                "retryableStatusCodes": ["UNAVAILABLE"],
            },
        }]
    })
    GRPC_OPTIONS = [('grpc.max_message_length', 100000000), ('grpc.max_send_message_length', 100000000),
                    ('grpc.max_receive_message_length', 100000000), ('grpc.enable_retries', 1),
                    ('grpc.service_config', SERVICE_CONFIG_JSON)]


class CONFIG_FILE:
    """
    Constants related to the config file
    """
    CSLE_HOME_ENV_PARAM = "CSLE_HOME"
    CONFIG_FILE_NAME = "config.json"
    PARSED_CONFIG: Union[None, Config] = None


class CONTAINER_IMAGES:
    """
    String constants representing container images names
    """
    CSLE_PREFIX = "csle_"
    BASE = "base"
    DERIVED = "derived"
    DOCKERHUB_USERNAME = "kimham"
    BLANK = "csle_blank"
    BLANK_1 = "csle_blank_1"
    SSH_1 = "csle_ssh_1"
    ROUTER_2 = "csle_router_2"
    SAMBA_2 = "csle_samba_2"
    HONEYPOT_1 = "csle_honeypot_1"
    FTP_1 = "csle_ftp_1"
    HACKER_KALI_1 = "csle_hacker_kali_1"
    SHELLSHOCK_1 = "csle_shellshock_1"
    SQL_INJECTION_1 = "csle_sql_injection_1"
    CVE_2010_0426_1 = "csle_cve_2010_0426_1"
    CVE_2015_1427_1 = "csle_cve_2015_1427_1"
    HONEYPOT_2 = "csle_honeypot_2"
    SAMBA_1 = "csle_samba_1"
    CVE_2015_3306_1 = "csle_cve_2015_3306_1"
    CVE_2015_5602_1 = "csle_cve_2015_5602_1"
    CVE_2016_10033_1 = "csle_cve_2016_10033_1"
    CLIENT_1 = "csle_client_1"
    KAFKA_1 = "csle_kafka_1"
    ELK_1 = "csle_elk_1"
    ROUTER_1 = "csle_router_1"
    TELNET_1 = "csle_telnet_1"
    SSH_2 = "csle_ssh_2"
    SSH_3 = "csle_ssh_3"
    TELNET_2 = "csle_telnet_2"
    TELNET_3 = "csle_telnet_3"
    FTP_2 = "csle_ftp_2"
    OVS_1 = "csle_ovs_1"
    RYU_1 = "csle_ryu_1"
    PENGINE_EXPLOIT_1 = "csle_pengine_exploit_1"
    CVE_2014_0160_1 = "csle_cve_2014_0160_1"
    SPARK_1 = "csle_spark_1"
    FIVE_G_CORE_1 = "csle_5g_core_1"
    CLOUD_RAN_CU_1 = "csle_cloud_ran_cu_1"
    CLOUD_RAN_DU_1 = "csle_cloud_ran_du_1"
    CLOUD_RAN_UE_1 = "csle_cloud_ran_ue_1"
    SNORT_IDS_IMAGES = [ROUTER_2]
    OVS_IMAGES = [OVS_1]
    SPARK_IMAGES = [SPARK_1]
    OSSEC_IDS_IMAGES = [HONEYPOT_1, HONEYPOT_2, PENGINE_EXPLOIT_1, ROUTER_1,
                        ROUTER_2, SSH_1, SSH_2, SSH_3, TELNET_1, TELNET_2, TELNET_3, SPARK_1]
    ROUTER_IMAGES = [ROUTER_1, ROUTER_2]
    HACKER_IMAGES = [HACKER_KALI_1]
    CLIENT_IMAGES = [CLIENT_1]
    CADVISOR = "cadvisor"
    PGADMIN = "pgadmin"
    GRAFANA = "grafana"
    FIVE_G_CORE_IMAGES = [FIVE_G_CORE_1]
    FIVE_G_CU_IMAGES = [CLOUD_RAN_CU_1]
    FIVE_G_DU_IMAGES = [CLOUD_RAN_DU_1]


class CONTAINER_OS:
    """
    String constants representing OS of different containers
    """
    SSH_1_OS = "ubuntu"
    ROUTER_2_OS = "ubuntu"
    SAMBA_2_OS = "debian"
    HONEYPOT_1_OS = "ubuntu"
    FTP_1_OS = "ubuntu"
    OVS_1_OS = "ubuntu"
    RYU_1_OS = "ubuntu"
    HACKER_KALI_1_OS = "kali"
    SHELLSHOCK_1_OS = "debian"
    SQL_INJECTION_1_OS = "debian"
    CVE_2010_0426_1_OS = "debian"
    CVE_2015_1427_1_OS = "debian"
    HONEYPOT_2_OS = "ubuntu"
    SAMBA_1_OS = "debian"
    CVE_2015_3306_1_OS = "debian"
    CVE_2015_5602_1_OS = "debian"
    CVE_2016_10033_1_OS = "debian"
    CLIENT_1_OS = "ubuntu"
    KAFKA_1_OS = "ubuntu"
    ELK_1_OS = "ubuntu"
    ROUTER_1_OS = "ubuntu"
    TELNET_1_OS = "ubuntu"
    SSH_2_OS = "ubuntu"
    SSH_3_OS = "ubuntu"
    TELNET_2_OS = "ubuntu"
    TELNET_3_OS = "ubuntu"
    FTP_2_OS = "ubuntu"
    PENGINE_EXPLOIT_1_OS = "ubuntu"
    CVE_2014_0160_1_OS = "debian"
    SPARK_1_OS = "ubuntu"
    FIVE_G_CORE_1_OS = "ubuntu"
    CLOUD_RAN_CU_1_OS = "ubuntu"
    CLOUD_RAN_DU_1_OS = "ubuntu"
    CLOUD_RAN_UE_1_OS = "ubuntu"


class RENDERING:
    """
    Rendering constants
    """
    RECT_SIZE = 200
    BLACK = (0, 0, 0)
    BLUE = (0, 0, 255)
    RED = (205, 55, 35)
    RED_ALPHA = (255, 0, 0, 255)
    GREEN = (0, 128, 0)
    GREEN_ALPHA = (0, 128, 0, 255)
    LIME = (0, 255, 0)
    BLUE_PURPLE = (102, 102, 153)
    # LIME = (0, 255, 0)
    BLACK_ALPHA = (0, 0, 0, 255)
    WHITE_ALPHA = (255, 255, 255, 255)
    RED_ALPHA = (128, 0, 0, 255)
    WHITE = (255, 255, 255)
    GREY = (220, 220, 220)
    RESOURCES_DIR = "resources"
    LINE_WIDTH = 1
    CAPTION = "csle"
    DEFAULT_WIDTH = 950
    DEFAULT_HEIGHT = 900
    TITLE = "csle"
    FIREWALL_SPRITE_NAME = "firewall.png"
    HACKER_SPRITE_NAME = "hacker.png"
    FLAG_SPRITE_NAME = "flag_1.png"
    LINK_COLORS = [(132, 87, 87), (153, 0, 153), (153, 0, 0), (204, 204, 255), (0, 102, 0), (102, 0, 102),
                   (153, 153, 0),
                   (128, 128, 128), (51, 153, 255), (0, 153, 153), (204, 255, 153), (255, 204, 153), (255, 153, 153),
                   (51, 51, 255), (255, 229, 204)]


class FIREWALL:
    """
    Firewall string constants
    """
    DROP = "DROP"
    ACCEPT = "ACCEPT"


class NETWORKING:
    """
    Networking string constants
    """
    ETH0 = "eth0"
    ETH1 = "eth1"
    ETH2 = "eth2"
    ETH3 = "eth3"
    ETH4 = "eth4"
    ETH5 = "eth5"
    ETH6 = "eth6"
    ETH7 = "eth7"
    ETH8 = "eth8"
    ETH9 = "eth9"
    ETH10 = "eth10"


class SERVICES:
    """
    Services constants
    """
    service_lookup = {}
    service_lookup["none"] = 0
    service_lookup["finger"] = 1
    service_lookup["mongo"] = 2
    service_lookup["mongod"] = 2
    service_lookup["tomcat"] = 3
    service_lookup["teamspeak"] = 4
    service_lookup["ts3"] = 4
    service_lookup["snmp"] = 5
    service_lookup["irc"] = 6
    service_lookup["ntp"] = 7
    service_lookup["postgres"] = 8
    service_lookup["postgresql"] = 8
    service_lookup["kafka"] = 9
    service_lookup["smtp"] = 10
    service_lookup["ssh"] = 11
    service_lookup["pengine"] = 12
    service_lookup["cassandra"] = 13
    service_lookup["telnet"] = 14
    service_lookup["http"] = 15
    service_lookup["http-proxy"] = 15
    service_lookup["gopher"] = 16
    service_lookup["kerberos"] = 17
    service_lookup["netbios"] = 18
    service_lookup["imap"] = 19
    service_lookup["dhcp"] = 20
    service_lookup["hdfs"] = 21
    service_lookup["netconf"] = 22
    service_lookup["dns"] = 23
    service_lookup["domain"] = 23
    service_lookup["mysql"] = 24
    service_lookup["docker"] = 25
    service_lookup["ventrilo"] = 26
    service_lookup["bittorrent"] = 27
    service_lookup["bitcoin"] = 28
    service_lookup["ftp"] = 29
    service_lookup["unknown"] = 30
    service_lookup["apani1"] = 31
    service_lookup["eforward"] = 32
    service_lookup["XmlIpcRegSvc"] = 33
    service_lookup["xmlipcregsvc"] = 33
    service_lookup["ajp13"] = 34
    service_lookup["wiegand"] = 35
    service_lookup["netiq-voipa"] = 36
    service_lookup["fmpro-v6"] = 37
    service_lookup["piccolo"] = 38
    service_lookup["dbdb"] = 39
    service_lookup["clariion-evr01"] = 40
    service_lookup["worldfusion2"] = 41
    service_lookup["esimport"] = 42
    service_lookup["ncdmirroring"] = 43
    service_lookup["abb-escp"] = 44
    service_lookup["directnet"] = 45
    service_lookup["fln - spx"] = 46
    service_lookup["netspeak-is"] = 47
    service_lookup["sec-pc2fax-srv"] = 48
    service_lookup["ridgeway2"] = 49
    service_lookup["fjicl-tep-b"] = 50
    service_lookup["ddt"] = 51
    service_lookup["informer"] = 52
    service_lookup["3m-image-lm"] = 53
    service_lookup["corelccam"] = 54
    service_lookup["plysrv-http"] = 56
    service_lookup["jdmn-port"] = 57
    service_lookup["evtp-data"] = 58
    service_lookup["can-ferret-ssl"] = 59
    service_lookup["efi-lm"] = 60
    service_lookup["landmarks"] = 61
    service_lookup["saris"] = 62
    service_lookup["powerguardian"] = 63
    service_lookup["sstp-1"] = 64
    service_lookup["escvpnet"] = 65
    service_lookup["mentaserver"] = 66
    service_lookup["nokia-ann-ch2"] = 67
    service_lookup["sip"] = 68
    service_lookup["mccwebsvr-port"] = 69
    service_lookup["newheights"] = 70
    service_lookup["lmp"] = 71
    service_lookup["vrml-multi-use"] = 71
    service_lookup["lotusnotes"] = 72
    service_lookup["dsmipv6"] = 73
    service_lookup["can-dch"] = 74
    service_lookup["hacl-monitor"] = 75
    service_lookup["spiral-admin"] = 76
    service_lookup["rapidmq-reg"] = 77
    service_lookup["neto-wol-server"] = 78
    service_lookup["pdb"] = 79
    service_lookup["directplay8"] = 80
    service_lookup["bis-web"] = 81
    service_lookup["senomix06"] = 82
    service_lookup["rsmtp"] = 83
    service_lookup["apc-9951"] = 84
    service_lookup["faxportwinport"] = 85
    service_lookup["mac-srvr-admin"] = 86
    service_lookup["vrts-at-port"] = 87
    service_lookup["vrtstrapserver"] = 88
    service_lookup["mtrgtrans"] = 89
    service_lookup["e-builder"] = 90
    service_lookup["ansoft-lm-1"] = 91
    service_lookup["ktelnet"] = 92
    service_lookup["pxc-ntfy"] = 93
    service_lookup["sybasesrvmon"] = 94
    service_lookup["opsmgr"] = 95
    service_lookup["fcp-srvr-inst2"] = 96
    service_lookup["itm-lm"] = 97
    service_lookup["ncconfig"] = 98
    service_lookup["client-ctrl"] = 99
    service_lookup["aairnet-2"] = 100
    service_lookup["servistaitsm"] = 101
    service_lookup["nfsrdma"] = 102
    service_lookup["cockroachdb"] = 103
    service_lookup["glassfish"] = 104
    service_lookup["samba"] = 105
    service_lookup["netbios-ssn"] = 106
    service_lookup["microsoft-ds"] = 107
    service_lookup["vrace"] = 108
    service_lookup["wap-wsp"] = 109
    service_lookup["elasticsearch"] = 110

    #
    service_lookup_inv = {v: k for k, v in service_lookup.items()}


class VULNERABILITIES:
    """
    Vulnerabilities constants
    """
    vuln_lookup = {}
    vuln_lookup["none"] = 0
    vuln_lookup["heartbleed"] = 1
    vuln_lookup["ghostcat"] = 2
    vuln_lookup["sql_injection"] = 3
    vuln_lookup["weak_password"] = 4
    vuln_lookup["drown"] = 5
    vuln_lookup["eternal_blue"] = 6
    vuln_lookup["shellshock"] = 7
    vuln_lookup["poodle"] = 8
    vuln_lookup["timthumb"] = 9
    vuln_lookup["CVE-2020-8620"] = 10
    vuln_lookup["CVE-2020-8617"] = 11
    vuln_lookup["CVE-2020-8616"] = 12
    vuln_lookup["CVE-2019-6470"] = 13
    vuln_lookup["CVE-2020-8623"] = 14
    vuln_lookup["CVE-2020-8621"] = 15
    vuln_lookup["CVE-2020-8624"] = 16
    vuln_lookup["CVE-2020-8622"] = 17
    vuln_lookup["CVE-2020-8619"] = 18
    vuln_lookup["CVE-2020-8618"] = 19
    vuln_lookup["CVE-2014-9278"] = 20
    vuln_lookup["ssh-weak-password"] = 21
    vuln_lookup["telnet-weak-password"] = 22
    vuln_lookup["ftp-weak-password"] = 23
    vuln_lookup["CVE-2020-15523"] = 24
    vuln_lookup["CVE-2020-14422"] = 25
    vuln_lookup["PACKETSTORM:157836"] = 26
    vuln_lookup["unknown"] = 27
    vuln_lookup_inv = {v: k for k, v in vuln_lookup.items()}
    default_cvss = 2.0


class OS:
    """
    Operating systems constants
    """
    os_lookup = {}
    os_lookup["unknown"] = 0
    os_lookup["windows"] = 1
    os_lookup["ubuntu"] = 2
    os_lookup["kali"] = 3
    os_lookup["suse"] = 4
    os_lookup["centos"] = 5
    os_lookup["fedora"] = 6
    os_lookup["debian"] = 7
    os_lookup["redhat"] = 8
    os_lookup["linux"] = 9
    os_lookup_inv = {v: k for k, v in os_lookup.items()}
    KALI = "kali"
    UBUNTU = "ubuntu"


class SECLISTS:
    """
    Constants related to seclists
    """
    TOP_USERNAMES_SHORTLIST = "/SecLists/Usernames/top-usernames-shortlist.txt"


class NMAP:
    """
    Constants related to nmap commands
    """
    SHELL_ESCAPE = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    RESULTS_DIR = "/home/agent/"
    SPEED_ARGS = "--min-rate 100000 --max-retries 1 -T5 -n"
    FILE_ARGS = "-oX"
    TELNET_BRUTE_SUBNET = "-p 23 --script telnet-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                          + ",passdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                          + ",telnet-brute.timeout=8s,brute.firstonly=true"
    TELNET_BRUTE_HOST = "-p 23 --script telnet-brute --script-args userdb=" \
                        + SECLISTS.TOP_USERNAMES_SHORTLIST \
                        + ",passdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                        + ",telnet-brute.timeout=8s,brute.firstonly=true"
    SSH_BRUTE_SUBNET = "-p 22 --script ssh-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                       + ",passdb=" \
                       + SECLISTS.TOP_USERNAMES_SHORTLIST + ",ssh-brute.timeout=8s,brute.firstonly=true"
    SSH_BRUTE_HOST = "-p 22 --script ssh-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                     + ",passdb=" \
                     + SECLISTS.TOP_USERNAMES_SHORTLIST + ",ssh-brute.timeout=8s,brute.firstonly=true"
    FTP_BRUTE_SUBNET = "-p 21 --script ftp-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                       + ",passdb=" \
                       + SECLISTS.TOP_USERNAMES_SHORTLIST + ",ftp-brute.timeout=8s,brute.firstonly=true"
    FTP_BRUTE_HOST = "-p 21 --script ftp-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                     + ",passdb=" \
                     + SECLISTS.TOP_USERNAMES_SHORTLIST + ",ftp-brute.timeout=8s,brute.firstonly=true"
    CASSANDRA_BRUTE_SUBNET = "-p 9160 --script cassandra-brute --script-args userdb=" \
                             + SECLISTS.TOP_USERNAMES_SHORTLIST + ",passdb=" \
                             + SECLISTS.TOP_USERNAMES_SHORTLIST + ",cassandra-brute.timeout=8s,brute.firstonly=true"
    CASSANDRA_BRUTE_HOST = "-p 9160 --script cassandra-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                           + ",passdb=" \
                           + SECLISTS.TOP_USERNAMES_SHORTLIST + ",cassandra-brute.timeout=8s,brute.firstonly=true"
    IRC_BRUTE_SUBNET = "-p 6667 --script irc-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                       + ",passdb=" \
                       + SECLISTS.TOP_USERNAMES_SHORTLIST + ",irc-brute.timeout=8s,brute.firstonly=true"
    IRC_BRUTE_HOST = "-p 6667 --script irc-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                     + ",passdb=" \
                     + SECLISTS.TOP_USERNAMES_SHORTLIST + ",irc-brute.timeout=8s,brute.firstonly=true"
    MONGO_BRUTE_SUBNET = "-p 27017 --script mongo-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                         + ",passdb=" \
                         + SECLISTS.TOP_USERNAMES_SHORTLIST + ",mongo-brute.timeout=8s,brute.firstonly=true"
    MONGO_BRUTE_HOST = "-p 27017 --script mongo-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                       + ",passdb=" \
                       + SECLISTS.TOP_USERNAMES_SHORTLIST + ",mongo-brute.timeout=8s,brute.firstonly=true"
    MYSQL_BRUTE_SUBNET = "-p 27017 --script mysql-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                         + ",passdb=" \
                         + SECLISTS.TOP_USERNAMES_SHORTLIST + ",mysql-brute.timeout=8s,brute.firstonly=true"
    MYSQL_BRUTE_HOST = "-p 27017 --script mysql-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                       + ",passdb=" \
                       + SECLISTS.TOP_USERNAMES_SHORTLIST + ",mysql-brute.timeout=8s,brute.firstonly=true"
    SMTP_BRUTE_SUBNET = "-p 25 --script smtp-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                        + ",passdb=" \
                        + SECLISTS.TOP_USERNAMES_SHORTLIST + ",smtp-brute.timeout=8s,brute.firstonly=true"
    SMTP_BRUTE_HOST = "-p 25 --script smtp-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                      + ",passdb=" \
                      + SECLISTS.TOP_USERNAMES_SHORTLIST + ",smtp-brute.timeout=8s,brute.firstonly=true"
    POSTGRES_BRUTE_SUBNET = "-p 5432 --script pgsql-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                            + ",passdb=" \
                            + SECLISTS.TOP_USERNAMES_SHORTLIST + ",pgsql-brute.timeout=8s,brute.firstonly=true"
    POSTGRES_BRUTE_HOST = "-p 5432 --script pgsql-brute --script-args userdb=" + SECLISTS.TOP_USERNAMES_SHORTLIST \
                          + ",passdb=" \
                          + SECLISTS.TOP_USERNAMES_SHORTLIST + ",pgsql-brute.timeout=8s,brute.firstonly=true"
    SAMBA_CVE_2017_7494_SCAN = "--script smb-vuln-cve-2017-7494 --script-args smb-vuln-cve-2017-7494.check-version " \
                               "-p445"
    FIREWALK_HOST = "--script=firewalk --traceroute --script-args=firewalk.max-retries=1,firewalk.probe-timeout=800ms"
    HTTP_ENUM = "--script=http-enum"
    HTTP_GREP = "--script=http-grep"
    FINGER = "--script=finger"


class AUXILLARY:
    """
    Auxillary constants
    """
    USER_PLACEHOLDER = "USER_PLACEHOLDER"
    PW_PLACEHOLDER = "USER_PLACEHOLDER"


class NMAP_XML:
    """
    Constants related to nmap XML output
    """
    HOST = "host"
    STATUS = "status"
    ADDRESS = "address"
    HOSTNAMES = "hostnames"
    PORTS = "ports"
    TRACE = "trace"
    HOP = "hop"
    OS = "os"
    STATE = "state"
    REASON = "reason"
    ARP_RESPONSE = "arp-response"
    STATUS_UP = "up"
    ADDR = "addr"
    ADDR_TYPE = "addrtype"
    HOSTNAME = "hostname"
    NAME = "name"
    PORT = "port"
    PORT_ID = "portid"
    UNKNOWN = "unknown"
    SERVICE = "service"
    SCRIPT = "script"
    OPEN_STATE = "open"
    OS_MATCH = "osmatch"
    ACCURACY = "accuracy"
    OS_CLASS = "osclass"
    VENDOR = "vendor"
    OS_FAMILY = "osfamily"
    ELEM = "elem"
    KEY = "key"
    CVSS = "cvss"
    ID = "id"
    TABLE = "table"
    IP = "ip"
    IPADDR = "ipaddr"
    TTL = "ttl"
    RTT = "rtt"
    HOST = "host"
    MAC = "mac"
    VULNERS_SCRIPT_ID = "vulners"
    TELNET_BRUTE_SCRIPT_ID = "telnet-brute"
    SSH_BRUTE_SCRIPT_ID = "ssh-brute"
    FTP_BRUTE_SCRIPT_ID = "ftp-brute"
    CASSANDRA_BRUTE_SCRIPT_ID = "cassandra-brute"
    IRC_BRUTE_SCRIPT_ID = "irc-brute"
    MONGO_BRUTE_SCRIPT_ID = "mongo-brute"
    MYSQL_BRUTE_SCRIPT_ID = "mysql-brute"
    SMTP_BRUTE_SCRIPT_ID = "smtp-brute"
    POSTGRES_BRUTE_SCRIPT_ID = "postgres-brute"
    BRUTE_SCRIPTS = [TELNET_BRUTE_SCRIPT_ID, SSH_BRUTE_SCRIPT_ID, FTP_BRUTE_SCRIPT_ID, CASSANDRA_BRUTE_SCRIPT_ID,
                     IRC_BRUTE_SCRIPT_ID, MONGO_BRUTE_SCRIPT_ID, MYSQL_BRUTE_SCRIPT_ID, SMTP_BRUTE_SCRIPT_ID,
                     POSTGRES_BRUTE_SCRIPT_ID]
    USERNAME = "username"
    PASSWORD = "password"
    ACCOUNTS = "Accounts"
    HTTP_ENUM_SCRIPT = "http-enum"
    OUTPUT = "output"
    HTTP_GREP_SCRIPT = "http-grep"
    VULSCAN_SCRIPT = "vulscan"
    VERSION = "version"
    SERVICEFP = "servicefp"


class SSH:
    """
    Constants related to the SSH service
    """
    SERVICE_NAME = "ssh"
    DEFAULT_PORT = 22
    DIRECT_CHANNEL = "direct-tcpip"
    MAX_FILE_READ_BYTES = 50000


class TELNET:
    """
    Constants related to the Telnet service
    """
    PROMPT = b':~$'
    LOCALHOST = "127.0.0.1"
    LOGIN_PROMPT = b"login: "
    PASSWORD_PROMPT = b"Password: "
    INCORRECT_LOGIN = "Login incorrect"
    SERVICE_NAME = "telnet"
    DEFAULT_PORT = 23


class FTP:
    """
    Constants related to the FTP service
    """
    INCORRECT_LOGIN = "Login incorrect"
    SERVICE_NAME = "ftp"
    DEFAULT_PORT = 21
    LOCALHOST = "127.0.0.1"
    LFTP_PROMPT = ":~>"
    LFTP_PROMPT_2 = ":/>"
    LFTP_PREFIX = "lftp ftp://"
    ACCESS_FAILED = "Access failed"


class HTTPS:
    """
    Constants related to the HTTPS service
    """
    SERVICE_NAME = "HTTPS"
    DEFAULT_PORT = 443
    OK_STATUS_CODE = 200
    UNAUTHORIZED_STATUS_CODE = 401
    CREATED_STATUS_CODE = 201
    BAD_REQUEST_STATUS_CODE = 400
    CONFLICT_STATUS_CODE = 409
    NOT_FOUND_STATUS_CODE = 404
    INTERNAL_SERVER_ERROR_STATUS_CODE = 500
    METHOD_NOT_ALLOWED_CODE = 405


class RETHINKDB:
    """
    Constants related to the RethinkDb service
    """
    SERVICE_NAME = "rethinkdb"
    DEFAULT_PORT = 28015


class COCKROACH:
    """
    Constants related to the Cockroach service
    """
    SERVICE_NAME = "cockroach"
    DEFAULT_PORT = 26257


class TOMCAT:
    """
    Constants related to the TOMCAT service
    """
    SERVICE_NAME = "tomcat"
    DEFAULT_PORT = 8080


class TEAMSPEAK3:
    """
    Constants related to the Teamspeak3 service
    """
    SERVICE_NAME = "teamspeak3"
    DEFAULT_PORT = 30033


class IRC:
    """
    Constants related to the IRC service
    """
    SERVICE_NAME = "irc"
    DEFAULT_PORT = 194


class POSTGRES:
    """
    Constants related to the Postgres service
    """
    SERVICE_NAME = "postgres"
    DEFAULT_PORT = 5432


class SMTP:
    """
    Constants related to the SMTP service
    """
    SERVICE_NAME = "smtp"
    DEFAULT_PORT = 25


class MYSQL:
    """
    Constants related to the MySQL service
    """
    SERVICE_NAME = "mysql"
    DEFAULT_PORT = 3306


class NTP:
    """
    Constants related to the NTP service
    """
    SERVICE_NAME = "ntp"
    DEFAULT_PORT = 123


class SNMP:
    """
    Constants related to the SNMP service
    """
    SERVICE_NAME = "snmp"
    DEFAULT_PORT = 161


class HTTP:
    """
    Constants related to the HTTP service
    """
    SERVICE_NAME = "http"
    DEFAULT_PORT = 80
    HTTP_PROTOCOL_PREFIX = "http://"
    DEFAULT_TIMEOUT = 5


class SPARK:
    """
    Constants related to the spark service
    """
    SERVICE_NAME = "spark"
    DEFAULT_PORT = 7077
    SPARK_PROTOCOL_PREFIX = "spark://"


class DNS:
    """
    Constants related to the DNS service
    """
    SERVICE_NAME = "dns"
    DEFAULT_PORT = 53


class MONGO:
    """
    Constants related to the MongoDB service
    """
    SERVICE_NAME = "mongo"
    DEFAULT_PORT = 27017


class CASSANDRA:
    """
    Constants related to the Cassandra service
    """
    SERVICE_NAME = "cassandra"
    DEFAULT_PORT = 9042


class SAMBA:
    """
    Constants related to the Samba service
    """
    SERVICE_NAME = "samba"
    USER = "sambacry"
    PW = "nosambanocry"
    BACKDOOR_USER = "ssh_backdoor_sambapwned"
    BACKDOOR_PW = "sambapwnedpw"
    PORT = 445
    ALREADY_EXISTS = "already exists"
    ERROR = "Error"
    AUTH_OK = "Authentication ok"
    VERIFYING = "Verifying"
    VULNERABILITY_NAME = "cve-2017-7494"


class CVE_2010_0426:
    """
    Constants related to CVE-2010-0426
    """
    SERVICE_NAME = "sudoedit"
    BACKDOOR_USER = "ssh_backdoor_cve10_0426pwn"
    BACKDOOR_PW = "cve_2010_0426_pwnedpw"
    EXPLOIT_FILE = "/etc/fstab"
    VULNERABILITY_NAME = "cve-2010-0426"


class CVE_2015_5602:
    """
    Constants related to CVE-2015-5602
    """
    SERVICE_NAME = "sudoedit"
    BACKDOOR_USER = "ssh_backdoor_cve15_5602pwn"
    BACKDOOR_PW = "cve_2015_5602_pwnedpw"
    ROOT_PW = "cve_2015_5602_temp_root_pw"
    VULNERABILITY_NAME = "cve-2015-5602"


class CVE_2015_3306:
    """
    Constants related to CVE-2015-3306
    """
    SERVICE_NAME = "proftpd"
    BACKDOOR_USER = "ssh_backdoor_cve2015_3306_pwned"
    BACKDOOR_PW = "cve2015_3306_pwnedpw"
    PORT = 21
    VULNERABILITY_NAME = "cve-2015-3306"


class CVE_2016_10033:
    """
    Constants related to CVE-2016-10033
    """
    SERVICE_NAME = "http"
    BACKDOOR_USER = "ssh_backdoor_2016_10033_pwn"
    BACKDOOR_PW = "cve_2016_10033_pwnedpw"
    PORT = 80
    VULNERABILITY_NAME = "cve-2016-10033"


class CVE_2015_1427:
    """
    Constants related to CVE-2015-1427
    """
    SERVICE_NAME = "elasticsearch"
    BACKDOOR_USER = "ssh_backdoor_cve_2015_1427_pwned"
    BACKDOOR_PW = "cve_2015_1427_pwnedpw"
    PORT = 9200
    VULNERABILITY_NAME = "cve-2015-1427"


class SHELLSHOCK:
    """
    Constants related to ShellShock
    """
    SERVICE_NAME = "http"
    BACKDOOR_USER = "ssh_backdoor_shellshocked"
    BACKDOOR_PW = "shellshockedpw"
    PORT = 80
    VULNERABILITY_NAME = "cve-2014-6271"


class DVWA_SQL_INJECTION:
    """
    Constants related to DVWA SQL Injection Vulnerabilities
    """
    SERVICE_NAME = "http"
    EXPLOIT_USER = "pablo"
    EXPLOIT_PW = "0d107d09f5bbe40cade3de5c71e9e9b7"
    EXPLOIT_OUTPUT_FILENAME = "dvwa_sql_injection_result.txt"
    PORT = 80
    VULNERABILITY_NAME = "dvwa_sql_injection"


class PENGINE_EXPLOIT:
    """
    Constants related to Pengine Exploit
    """
    SERVICE_NAME = "http"
    PORT = 4000
    VULNERABILITY_NAME = "pengine-exploit"
    BACKDOOR_USER = "ssh_backdoor_pengine_exploitpwn"
    BACKDOOR_PW = "ssh_backdoor_pengine_exploitpwnpw"


class COMMON:
    """
    Common constants
    """
    CVE_FILE = "/allitems_prep.csv"
    SERVICES_FILE = "/nmap-services"
    DEFAULT_RECV_SIZE = 5000
    LARGE_RECV_SIZE = 1000000
    FLAG_FILENAME_PREFIX = "flag"
    LOCALHOST = "localhost"
    LOCALHOST_127_0_0_1 = "127.0.0.1"
    LOCALHOST_127_0_1_1 = "127.0.1.1"


class COMMANDS:
    """
    Constants related to arbitrary commands
    """
    CHANNEL_WHOAMI = "whoami\n"
    BASH = "bash"
    TAIL = "tail"
    CHANNEL_SU_ROOT = "su root\n"
    CHANNEL_ROOT = "root\n"
    LIST_CACHE = "ls -1 "
    LS = "ls"
    SUDO = "sudo"
    CHMOD_777 = "chmod 777"
    CHMOD_U_RWX = "chmod u+rwx"
    SLASH_DELIM = "/"
    COLON_DELIM = ":"
    NEW_LINE_DELIM = "\n"
    DASH_DELIM = "-"
    UNDERSCORE_DELIM = "_"
    STAR_DELIM = "*"
    DOT_DELIM = "."
    PIPE_DELIM = "|"
    TOUCH = "touch"
    NOHUP = "nohup"
    AMP = "&"
    PKILL = "pkill -f"
    RM_F = "rm -f"
    SPACE_DELIM = " "
    TMP_DIR = "tmp"
    ROOT_DIR = "root"
    HOME_DIR = "home"
    SUDO_RM_RF = "sudo rm -rf"
    SUDO_TOUCH = "sudo touch"
    ECHO = "echo"
    LS_HOME = "ls /home/"
    RM_F_HOME = "rm -f home/"
    SUDO_ADD_ROUTE = "sudo route add"
    NETMASK = "netmask"
    CLEAR_IPTABLES = "sudo iptables -F"
    IPTABLES_APPEND_INPUT = "sudo iptables -A INPUT"
    IPTABLES_APPEND_OUTPUT = "sudo iptables -A OUTPUT"
    IPTABLES_APPEND_FORWARD = "sudo iptables -A FORWARD"
    ARPTABLES_APPEND_INPUT = "sudo arptables -A INPUT"
    ARPTABLES_APPEND_OUTPUT = "sudo arptables -A OUTPUT"
    ARPTABLES_APPEND_FORWARD = "sudo arptables -A FORWARD"
    CHANGE_PERMISSION_LOG_DIRS = "sudo chmod -R 777 /var"
    UPDATE_RULESET = "/pulledpork/pulledpork.pl -c /pulledpork/etc/pulledpork.conf -l -P -E -H SIGHUP"
    SNORT_PID = "/var/run//snort_eth1:eth0.pid"
    SNORT_DUMP_STATS = "kill -SIGUSR1 {}"
    SNORT_ROTATE_STATS = "kill -SIGUSR2 {}"
    PS_AUX = "ps -aux"
    PS_AXR = "ps -axr"
    GREP = "grep"
    START_CLIENT_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /client_manager.py --port {} --logdir {} " \
                           "--logfile {} --maxworkers {} &"
    SEARCH_CLIENT_MANAGER = "/root/miniconda3/bin/python3 /client_manager.py"
    START_KAFKA_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /kafka_manager.py --port {} --logdir {} " \
                          "--logfile {} --maxworkers {} &"
    START_ELK_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /elk_manager.py --port {} --logdir {} --logfile {} " \
                        "--maxworkers {} &"
    START_RYU_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /ryu_manager.py --port {} --logdir {} --logfile {} " \
                        "--maxworkers {} &"
    START_SNORT_IDS_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /snort_ids_manager.py --port {} --logdir {} " \
                              "--logfile {} --maxworkers {} &"
    START_OSSEC_IDS_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /ossec_ids_manager.py --port {} --logdir {} " \
                              "--logfile {} --maxworkers {} &"
    START_HOST_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /host_manager.py --port {} --logdir {} " \
                         "--logfile {} --maxworkers {} &"
    START_TRAFFIC_MANAGER = "sudo nohup /root/miniconda3/bin/python3 /traffic_manager.py --port {} --logdir {} " \
                            "--logfile {} --maxworkers {} &"
    START_FIVE_G_CORE_MANAGER = ("sudo nohup /root/miniconda3/bin/python3 /five_g_core_manager.py --port {} "
                                 "--logdir {} --logfile {} --maxworkers {} &")
    START_FIVE_G_CU_MANAGER = ("sudo nohup /root/miniconda3/bin/python3 /five_g_cu_manager.py --port {} "
                               "--logdir {} --logfile {} --maxworkers {} &")
    START_FIVE_G_DU_MANAGER = ("sudo nohup /root/miniconda3/bin/python3 /five_g_du_manager.py --port {} "
                               "--logdir {} --logfile {} --maxworkers {} &")
    START_SDN_CONTROLLER = "sudo nohup /root/miniconda3/bin/python3 /ryu_controller.py --port {} --webport {} " \
                           "--controller {} &"
    SEARCH_KAFKA_MANAGER = "/root/miniconda3/bin/python3 /kafka_manager.py"
    SEARCH_ELK_MANAGER = "/root/miniconda3/bin/python3 /elk_manager.py"
    SEARCH_RYU_MANAGER = "/root/miniconda3/bin/python3 /ryu_manager.py"
    SEARCH_SNORT_IDS_MANAGER = "/root/miniconda3/bin/python3 /snort_ids_manager.py"
    SEARCH_OSSEC_IDS_MANAGER = "/root/miniconda3/bin/python3 /ossec_ids_manager.py"
    SEARCH_HOST_MANAGER = "/root/miniconda3/bin/python3 /host_manager.py"
    SEARCH_TRAFFIC_MANAGER = "/root/miniconda3/bin/python3 /traffic_manager.py"
    SEARCH_SDN_CONTROLLER = "/root/miniconda3/bin/python3 /ryu_controller.py"
    DOCKER_STATS_MANAGER_PIDFILE = "/var/log/csle/statsmanager.pid"
    DOCKER_STATS_MANAGER_OUTFILE = "/var/log/csle/statsmanager.out"
    START_DOCKER_STATS_MANAGER = "nohup csle statsmanager {} " \
                                 f"& > {DOCKER_STATS_MANAGER_OUTFILE} " \
                                 f"&& echo $! > {DOCKER_STATS_MANAGER_PIDFILE}"
    CLUSTER_MANAGER_PIDFILE = "/var/log/csle/clustermanager.pid"
    CLUSTER_MANAGER_OUTFILE = "/var/log/csle/clustermanager.out"
    START_CLUSTER_MANAGER = "nohup csle clustermanager {} " \
                            f"& > {CLUSTER_MANAGER_OUTFILE} " \
                            f"&& echo $! > {CLUSTER_MANAGER_PIDFILE}"
    SEARCH_DOCKER_STATS_MANAGER = "statsmanager"
    SEARCH_PROMETHEUS = "prometheus"
    PROMETHEUS_PID_FILE = "/var/log/csle/prometheus.pid"
    PROMETHEUS_LOG_FILE = "/var/log/csle/prometheus.log"
    PROMETHEUS_CONFIG_FILE = f"${CONFIG_FILE.CSLE_HOME_ENV_PARAM}/management-system/prometheus/prometheus.yml"
    POSTGRESQL_LOG_DIR = "/var/log/postgresql/"
    NGINX_LOG_DIR = "/var/log/nginx/"
    PROMETHEUS_PORT = 9090
    START_PROMETHEUS = f"nohup $CSLE_HOME/management-system/prometheus/prometheus " \
                       f"--config.file={PROMETHEUS_CONFIG_FILE} " \
                       "--storage.tsdb.retention.size=10GB " \
                       f"--storage.tsdb.retention.time=5d & > {PROMETHEUS_LOG_FILE} " \
                       f"&& echo $! > {PROMETHEUS_PID_FILE}"
    SEARCH_NODE_EXPORTER = "node_exporter"
    SEARCH_MONITOR = "server.py"
    NODE_EXPORTER_PORT = 9100
    GRAFANA_PORT = 3000
    MANAGEMENT_SYSTEM_PORT = 7777
    FLASK_PORT = 7777
    START_GRAFANA = f"docker run -d -p {GRAFANA_PORT}:{GRAFANA_PORT} --name grafana grafana/grafana"
    CADVISOR_PORT = 8080
    PGADMIN_PORT = 7778
    DOCKER_ENGINE_PORT = 2375
    START_CADVISOR = "docker run  -dt --volume=/:/rootfs:ro   --volume=/var/run:/var/run:ro   " \
                     "--volume=/sys:/sys:ro   " \
                     "--volume=/var/lib/docker/:/var/lib/docker:ro   --volume=/dev/disk/:/dev/disk:ro   " \
                     f"--publish={CADVISOR_PORT}:{CADVISOR_PORT}  --name=cadvisor  " \
                     "gcr.io/cadvisor/cadvisor"
    PGADMIN_USERNAME = "csle@csle.com"
    PGADMIN_PW = "cslePassword"
    GRAFANA_USERNAME = "admin"
    GRAFANA_PW = "admin"
    START_PGADMIN = f"docker run -p 7778:80 -e 'PGADMIN_DEFAULT_EMAIL={PGADMIN_USERNAME}' -e " \
                    f"'PGADMIN_DEFAULT_PASSWORD={PGADMIN_PW}' -d --name=pgadmin dpage/pgadmin4"
    CONTAINER_LOGS = "docker logs {}"
    CADVISOR_LOGS = "docker logs cadvisor"
    PGADMIN_LOGS = "docker logs pgadmin"
    GRAFANA_LOGS = "docker logs grafana"
    DOCKER_ENGINE_LOGS = "sudo -n /usr/bin/journalctl -u docker.service -n 100 --no-pager -e"
    DOCKER_ENGINE_LOGS_ALTERNATIVE = "sudo -n /bin/journalctl -u docker.service -n 100 --no-pager -e"
    NODE_EXPORTER_PID_FILE = "/var/log/csle/node_exporter.pid"
    CSLE_MGMT_WEBAPP_PID_FILE = "/var/log/csle/csle_mgmt_webapp.pid"
    NODE_EXPORTER_LOG_FILE = "/var/log/csle/node_exporter.log"
    FLASK_LOG_FILE = "/var/log/csle/flask.log"
    START_NODE_EXPORTER = f"nohup $CSLE_HOME/management-system/node_exporter/node_exporter " \
                          f"& > {NODE_EXPORTER_LOG_FILE} && echo $! " \
                          f"> {NODE_EXPORTER_PID_FILE}"
    BUILD_CSLE_MGMT_WEBAPP = f"cd ${CONFIG_FILE.CSLE_HOME_ENV_PARAM}/management-system/csle-mgmt-webapp && npm run " \
                             f"build"
    START_CSLE_MGMT_WEBAPP = f"nohup python ${CONFIG_FILE.CSLE_HOME_ENV_PARAM}/management-system/csle-mgmt-webapp" \
                             f"/server/server.py & > {FLASK_LOG_FILE}"
    GET_LATEST_PID = "$!"
    SAVE_PID = "echo {} > {}"
    KILL_PROCESS = "kill -9 {}"
    START_TRAINING_JOB = "nohup csle trainingjob {} &"
    START_SYSTEM_IDENTIFICATION_JOB = "nohup csle systemidentificationjob {} &"
    DOCKER_EXEC_COMMAND = "docker exec"
    PING = "ping"
    NGINX_STATUS = "service nginx status"
    POSTGRESQL_STATUS = "service postgresql status"
    POSTGRESQL_STATUS_VERSION = "service postgresql@15-main status"
    DOCKER_ENGINE_STATUS = "service docker status"
    POSTGRESQL_START = "sudo -n service postgresql start"
    POSTGRESQL_STOP = "sudo -n service postgresql stop"
    NGINX_START = "sudo -n service nginx start"
    NGINX_STOP = "sudo -n service nginx stop"
    DOCKER_ENGINE_START = "sudo -n service docker start"
    DOCKER_ENGINE_STOP = "sudo -n service docker stop"


class OVS:
    """
    String constants related to OVS
    """
    DEFAULT_BRIDGE_NAME = "ovs-br0"
    OVS_VSCTL = "ovs-vsctl"
    ADD_BR = "add-br"
    DEL_BR = "del-br"
    OVS_DOCKER = "ovs-docker"
    ADD_PORT = "add-port"
    IPADDRESS = "--ipaddress"
    SET_VLAN = "set-vlan"
    ADD_VETH_PEER_LINK = "ip link add dev {} type veth peer name {}"
    SET_INTERFACE = "set interface"
    TYPE_PATCH = "type=patch"
    OPTIONS_PEER = "options:peer"
    DELETE_VETH_PEER_LINK = "ip link delete {}"


class ETC_HOSTS:
    """
    Constants related to /etc/hosts configuration
    """
    DEFAULT_HOST_LINE_1 = "'127.0.0.1 localhost'"
    DEFAULT_HOST_LINE_2 = ":':1 localhost ip6-localhost ip6-loopback'"
    DEFAULT_HOST_LINE_3 = "'fe00::0 ip6-localnet'"
    DEFAULT_HOST_LINE_4 = "'ff00::0 ip6-mcastprefix'"
    DEFAULT_HOST_LINE_5 = "'ff02::1 ip6-allnodes'"
    DEFAULT_HOST_LINE_6 = "''ff02::2 ip6-allrouters''"
    APPEND_TO_ETC_HOSTS = "sudo -n tee -a /etc/hosts"


class FILE_PATTERNS:
    """
    Constants related to file patterns for parsing
    """
    COST_FILE_SUFFIX = "_cost.txt"
    NMAP_ACTION_RESULT_SUFFIX = ".xml"
    ALERTS_FILE_SUFFIX = "_alerts.txt"
    TXT_FILE_SUFFIX = ".txt"
    XML_FILE_SUFFIX = ".xm"
    CSV_SUFFIX = ".csv"
    LOG_SUFFIX = ".log"
    GZ_SUFFIX = ".gz"


class NIKTO:
    """
    Constants related to Nikto commands
    """
    BASE_ARGS = "-port 80 -Format xml --maxtime 60s -timeout 5 "
    HOST_ARG = "-h "
    OUTPUT_ARG = "-output "


class NIKTO_XML:
    """
    Constants related to Nikto XML parsing
    """
    NIKTOSCAN = "niktoscan"
    SCANDETAILS = "scandetails"
    ITEM = "item"
    ITEM_ID = "id"
    OSVDB_ID = "osvdbid"
    DESCR = "description"
    NAMELINK = "namelink"
    IPLINK = "iplink"
    URI = "uri"
    TARGETPORT = "targetport"
    TARGETIP = "targetip"
    SITENAME = "sitename"
    METHOD = "method"


class MASSCAN:
    """
    Constants related to Masscan commands
    """
    BASE_ARGS = "-p0-1024 --max-rate 100000 --max-retries 1 --wait 0"
    HOST_ARG = "--source-ip "
    OUTPUT_ARG = "-oX "


class SSH_BACKDOOR:
    """
    Constants related to creation of SSH backdoors
    """
    BACKDOOR_PREFIX = "ssh_backdoor"
    DEFAULT_PW = "csle"


class SHELL:
    """
    Constants related to shell commands
    """
    LIST_ALL_USERS = "cut -d: -f1 /etc/passwd"
    CHECK_FOR_SECLISTS = "test -e /SecLists && echo file exists || echo file not found"
    SAMBA_EXPLOIT = "/samba_exploit.py -e /libbindshell-samba.so -s data -r /data/libbindshell-samba.so -u " \
                    "sambacry -p nosambanocry -P 6699 -t "


class EXPLOIT_VULNERABILITES:
    """
    Constants related to exploit vulnerabilities
    """
    SSH_DICT_SAME_USER_PASS = "ssh-weak-password"
    FTP_DICT_SAME_USER_PASS = "ftp-weak-password"
    TELNET_DICTS_SAME_USER_PASS = "telnet-weak-password"
    IRC_DICTS_SAME_USER_PASS = "irc-weak-password"
    POSTGRES_DICTS_SAME_USER_PASS = "postgres-weak-password"
    SMTP_DICTS_SAME_USER_PASS = "smtp-weak-password"
    MYSQL_DICTS_SAME_USER_PASS = "mysql-weak-password"
    MONGO_DICTS_SAME_USER_PASS = "mongo-weak-password"
    CASSANDRA_DICTS_SAME_USER_PASS = "cassandra-weak-password"
    WEAK_PW_VULNS = [SSH_DICT_SAME_USER_PASS, FTP_DICT_SAME_USER_PASS, TELNET_DICTS_SAME_USER_PASS,
                     IRC_DICTS_SAME_USER_PASS, POSTGRES_DICTS_SAME_USER_PASS, SMTP_DICTS_SAME_USER_PASS,
                     MYSQL_DICTS_SAME_USER_PASS, MONGO_DICTS_SAME_USER_PASS, CASSANDRA_DICTS_SAME_USER_PASS]
    SAMBACRY_EXPLOIT = "cve-2017-7494"
    SHELLSHOCK_EXPLOIT = "cve-2014-6271"
    DVWA_SQL_INJECTION = "dvwa_sql_injection"
    CVE_2015_3306 = "cve-2015-3306"
    CVE_2015_1427 = "cve-2015-1427"
    CVE_2016_10033 = "cve-2016-10033"
    CVE_2010_0426 = "cve-2010-0426"
    CVE_2015_5602 = "cve-2015-5602"
    PENGINE_EXPLOIT = "pengine-exploit"
    CVE_2014_0160 = "cve-2014-0160"
    CVE_VULNS = [SAMBACRY_EXPLOIT, SHELLSHOCK_EXPLOIT, CVE_2015_3306, CVE_2015_1427, CVE_2016_10033, CVE_2010_0426,
                 CVE_2015_5602]
    PRIVILEGE_ESC_VULNS = [CVE_2010_0426, CVE_2015_5602]
    UNKNOWN = "unknown"
    WEAK_PASSWORD_CVSS = 10.0
    SAMBACRY_CVSS = 9.8
    SHELLSHOCK_CVSS = 9.8
    DVWA_SQL_INJECTION_CVSS = 9.5
    CVE_2015_3306_CVSS = 9.8
    CVE_2016_10033_CVSS = 9.8
    CVE_2010_0426_CVSS = 6
    CVE_2015_5602_CVSS = 6
    CVE_2015_1427_CVSS = 9.8
    PENGINE_EXPLOIT_CVSS = 9.8


class SUB_PROC_ENV:
    """
    Constants related to creation of Sub-proc-env environments
    """
    SLEEP_TIME_STARTUP = 5


class DUMMY_VEC_ENV:
    """
    Constants related to creation of Sub-proc-env environments
    """
    SLEEP_TIME_STARTUP = 1


class TRAFFIC_COMMANDS:
    """
    Constants related to traffic commands
    """
    TRAFFIC_GENERATOR_FILE_NAME = "traffic_generator.sh"
    BASH_PREAMBLE = "#!/bin/bash"
    CLIENT_MANAGER_FILE_NAME = "client_manager.py"
    KAFKA_MANAGER_FILE_NAME = "kafka_manager.py"
    ELK_MANAGER_FILE_NAME = "elk_manager.py"
    RYU_MANAGER_FILE_NAME = "ryu_manager.py"
    SNORT_IDS_MANAGER_FILE_NAME = "snort_ids_manager.py"
    OSSEC_IDS_MANAGER_FILE_NAME = "ossec_ids_manager.py"
    FIVE_G_CORE_MANAGER_FILE_NAME = "five_g_core_manager.py"
    FIVE_G_CU_MANAGER_FILE_NAME = "five_g_cu_manager.py"
    FIVE_G_DU_MANAGER_FILE_NAME = "five_g_du_manager.py"
    HOST_MANAGER_FILE_NAME = "host_manager.py"
    TRAFFIC_MANAGER_FILE_NAME = "traffic_manager.py"
    SDN_CONTROLLER_FILE_NAME = "ryu_controller.py"
    GENERIC_COMMANDS = "generic_commands"
    CLIENT_1_SUBNET = "client_1_subnet"
    DEFAULT_COMMANDS = {
        f"{CONTAINER_IMAGES.FTP_1}": [
            "timeout 5 ftp {} > /dev/null 2>&1",
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:8080 > /dev/null 2>&1"
        ],
        f"{CONTAINER_IMAGES.SSH_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:80 > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.TELNET_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {} > /dev/null 2>&1",
            "(sleep 2; echo test; sleep 2; echo test; sleep 3;) | telnet {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.HONEYPOT_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 snmpwalk -v2c {} -c csle_1234 > /dev/null 2>&1",
            "timeout 10 /irc_login_test.sh {} > /dev/null 2>&1",
            "timeout 5 psql -h {} -p 5432 > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SAMBA_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "(sleep 2; echo testcsleuser; sleep 3;) | smbclient -L {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SAMBA_2}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "(sleep 2; echo testcsleuser; sleep 3;) | smbclient -L {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SHELLSHOCK_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {} > /dev/null 2>&1",
            "timeout 5 snmpwalk -v2c {} -c csle_1234 > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SQL_INJECTION_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}/login.php > /dev/null 2>&1",
            "timeout 10 /irc_login_test.sh {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.CVE_2010_0426_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:8080 > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.CVE_2015_1427_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "snmpwalk -v2c {} -c csle_1234"],
        f"{CONTAINER_IMAGES.CVE_2015_3306_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "snmpwalk -v2c {} -c csle_1234",
            "timeout 5 curl {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.CVE_2015_5602_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.CVE_2016_10033_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.HONEYPOT_2}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 snmpwalk -v2c {} -c csle_1234 > /dev/null 2>&1",
            "timeout 10 /irc_login_test.sh {} > /dev/null 2>&1",
            "timeout 5 psql -h {} -p 5432 > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SSH_2}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 nslookup kim-hammar.com {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SSH_3}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 10 /irc_login_test.sh {} > /dev/null 2>&1"
        ],
        f"{CONTAINER_IMAGES.TELNET_2}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:8080 > /dev/null 2>&1",
            "(sleep 2; echo test; sleep 2; echo test; sleep 3;) | telnet {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.TELNET_3}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:8080 > /dev/null 2>&1",
            "(sleep 2; echo test; sleep 2; echo test; sleep 3;) | telnet {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.FTP_2}": [
            "timeout 5 ftp {} > /dev/null 2>&1",
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "(sleep 2; echo test; sleep 2; echo test; sleep 3;) | telnet {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.ROUTER_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.ROUTER_2}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.OVS_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1"],
        f"{CLIENT_1_SUBNET}": [
            # "sudo nmap -sS -p- " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -sP " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -sU -p- " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -sT -p- " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -sF -p- " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -sN -p- " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -sX -p- " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap -O --osscan-guess --max-os-tries 1 " + NMAP.SPEED_ARGS + " --host-timeout 5 {}
            # > /dev/null 2>&1",
            # "sudo nmap " + NMAP.HTTP_GREP + " " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1",
            # "sudo nmap " + NMAP.FINGER + " " + NMAP.SPEED_ARGS + " --host-timeout 5 {} > /dev/null 2>&1"
        ],
        f"{GENERIC_COMMANDS}": [
            "timeout 5 ping {} > /dev/null 2>&1",
            "timeout 5 traceroute {} > /dev/null 2>&1"
        ],
        f"{CONTAINER_IMAGES.PENGINE_EXPLOIT_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:4000 > /dev/null 2>&1",
            "timeout 5 curl --header \"Content-Type: application/json\" --request POST \
                 --data $'{{\"application\": \"pengine_sandbox\", \"ask\": "
            "\"problem(1, Rows), sudoku(Rows)\", \"chunk\": 1, \"destroy\": true, \"format\":\"json\", "
            "\"src_text\": \"problem(1, [[_,_,_,_,_,_,_,_,_],[_,_,_,_,_,3,_,8,5],[_,_,1,_,2,_,_,_,_],"
            "[_,_,_,5,_,7,_,_,_],[_,_,4,_,_,_,1,_,_],[_,9,_,_,_,_,_,_,_],[5,_,_,_,_,_,_,7,3],"
            "[_,_,2,_,1,_,_,_,_],[_,_,_,_,4,_,_,_,9]]).\n\"}}' {}"
        ],
        f"{CONTAINER_IMAGES.CVE_2014_0160_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:443 > /dev/null 2>&1"],
        f"{CONTAINER_IMAGES.SPARK_1}": [
            "timeout 5 sshpass -p 'testcsleuser' ssh -oStrictHostKeyChecking=no {} > /dev/null 2>&1",
            "timeout 5 curl {}:8080 > /dev/null 2>&1",
            "timeout 5 curl {}:8081 > /dev/null 2>&1",
            "/root/miniconda3/bin/python3 /spark_job.py --sparkmaster {} > /dev/null 2>&1"
        ]
    }


class AGENT:
    """
    String constants related to the agent
    """
    USER = "agent"
    PW = "agent"


class CITUS:
    """
    Constants related to CITUS
    """
    COORDINATOR_PORT = 5432


class CLUSTER_CONFIG:
    """
    Constants related to the cluster configuration
    """
    LEADER = False
    IP = "127.0.0.1"


class CSLE_ADMIN:
    """
    Constants related to the default csle admin account.
    """
    SSH_USER = "csle_admin"  # Should not be a simple user/pw, the automated attacker will figure it out
    SSH_PW = "csle@admin-pw_191"
    MANAGEMENT_USER = "admin"
    MANAGEMENT_PW = "admin"
    MANAGEMENT_FIRST_NAME = "admin"
    MANAGEMENT_LAST_NAME = "adminson"
    MANAGEMENT_ORGANIZATION = "CSLE"
    MANAGEMENT_EMAIL = "admin@csle.com"


class CSLE_GUEST:
    """
    Constants related to the default csle guest account
    """
    USER = "csle_guest"  # Should not be a simple user/pw combination, the automated attacker will figure it out
    PW = "csle@guest-pw_191"
    MANAGEMENT_USER = "guest"
    MANAGEMENT_PW = "guest"
    MANAGEMENT_FIRST_NAME = "guest"
    MANAGEMENT_LAST_NAME = "guestson"
    MANAGEMENT_ORGANIZATION = "CSLE"
    MANAGEMENT_EMAIL = "guest@csle.com"


class SYSTEM_IDENTIFICATION:
    """
    Constants related to the system identification process
    """
    NETWORK_CONF_FILE = "network_conf.pickle"
    DEFENDER_DYNAMICS_MODEL_FILE = "defender_dynamics_model.json"
    SIMULATION_TRACES_FILE = "simulation_taus.json"
    EMULATION_TRACES_FILE = "emulation_traces.json"
    SYSTEM_ID_LOGS_FILE = "system_id_log.csv"
    INTRUSION_CONDITIONAL = "intrusion"
    NO_INTRUSION_CONDITIONAL = "no_intrusion"


class AUXILLARY_COMMANDS:
    """
    Constants related to auxillary shell commands
    """
    WHOAMI = "whoami"


class INFO_DICT:
    """
    Constants for strings in the info dict of the csle_CTF Environment
    """
    EPISODE_LENGTH = "episode_length"
    FLAGS = "flags"
    INTRUSION_STATE = "intrusion_state"
    EARLY_STOPPED = "early_stopped"
    DEFENDER_STOPS_REMAINING = "defender_stops_remaining"
    DEFENDER_FIRST_STOP_STEP = "defender_first_stop_step"
    DEFENDER_SECOND_STOP_STEP = "defender_second_stop_step"
    DEFENDER_THIRD_STOP_STEP = "defender_third_stop_step"
    DEFENDER_FOURTH_STOP_STEP = "defender_fourth_stop_step"
    SUCCESSFUL_INTRUSION = "successful_intrusion"
    SNORT_SEVERE_BASELINE_REWARD = "snort_severe_baseline_reward"
    SNORT_WARNING_BASELINE_REWARD = "snort_warning_baseline_reward"
    SNORT_CRITICAL_BASELINE_REWARD = "snort_critical_baseline_reward"
    VAR_LOG_BASELINE_REWARD = "var_log_baseline_reward"
    STEP_BASELINE_REWARD = "step_baseline_reward"
    SNORT_SEVERE_BASELINE_STEP = "snort_severe_baseline_step"
    SNORT_WARNING_BASELINE_STEP = "snort_warning_baseline_step"
    SNORT_CRITICAL_BASELINE_STEP = "snort_critical_baseline_step"
    VAR_LOG_BASELINE_STEP = "var_log_baseline_step"
    STEP_BASELINE_STEP = "step_baseline_step"
    SNORT_SEVERE_BASELINE_CAUGHT_ATTACKER = "snort_severe_baseline_caught_attacker"
    SNORT_WARNING_BASELINE_CAUGHT_ATTACKER = "snort_warning_baseline_caught_attacker"
    SNORT_CRITICAL_BASELINE_CAUGHT_ATTACKER = "snort_critical_baseline_caught_attacker"
    VAR_LOG_BASELINE_CAUGHT_ATTACKER = "var_log_baseline_caught_attacker"
    STEP_BASELINE_CAUGHT_ATTACKER = "step_baseline_caught_attacker"
    SNORT_SEVERE_BASELINE_EARLY_STOPPING = "snort_severe_baseline_early_stopping"
    SNORT_WARNING_BASELINE_EARLY_STOPPING = "snort_warning_baseline_early_stopping"
    SNORT_CRITICAL_BASELINE_EARLY_STOPPING = "snort_critical_baseline_early_stopping"
    VAR_LOG_BASELINE_EARLY_STOPPING = "var_log_baseline_early_stopping"
    STEP_BASELINE_EARLY_STOPPING = "step_baseline_early_stopping"
    SNORT_SEVERE_BASELINE_UNCAUGHT_INTRUSION_STEPS = "snort_severe_baseline_uncaught_intrusion_steps"
    SNORT_WARNING_BASELINE_UNCAUGHT_INTRUSION_STEPS = "snort_warning_baseline_uncaught_intrusion_steps"
    SNORT_CRITICAL_BASELINE_UNCAUGHT_INTRUSION_STEPS = "snort_critical_baseline_uncaught_intrusion_steps"
    VAR_LOG_BASELINE_UNCAUGHT_INTRUSION_STEPS = "var_log_baseline_uncaught_intrusion_steps"
    STEP_BASELINE_UNCAUGHT_INTRUSION_STEPS = "step_baseline_uncaught_intrusion_steps"
    ATTACKER_COST = "attacker_cost"
    ATTACKER_COST_NORM = "attacker_cost_norm"
    ATTACKER_ALERTS = "attacker_alerts"
    ATTACKER_ALERTS_NORM = "attacker_alerts_norm"
    INTRUSION_STEP = "intrusion_step"
    UNCAUGHT_INTRUSION_STEPS = "uncaught_intrusion_steps"
    OPTIMAL_DEFENDER_REWARD = "optimal_defender_reward"
    ATTACKER_NON_LEGAL_ACTIONS = "attacker_non_legal_actions"
    DEFENDER_NON_LEGAL_ACTIONS = "defender_non_legal_actions"
    IDX = "idx"
    NON_LEGAL_ACTIONS = "non_legal_actions"
    ATTACKER_ACTION = "attacker_action"
    SNORT_SEVERE_BASELINE_FIRST_STOP_STEP = "snort_severe_baseline_first_stop_step"
    SNORT_WARNING_BASELINE_FIRST_STOP_STEP = "snort_warning_baseline_first_stop_step"
    SNORT_CRITICAL_BASELINE_FIRST_STOP_STEP = "snort_critical_baseline_first_stop_step"
    VAR_LOG_BASELINE_FIRST_STOP_STEP = "var_log_baseline_first_stop_step"
    STEP_BASELINE_FIRST_STOP_STEP = "step_baseline_uncaught_first_stop_step"
    SNORT_SEVERE_BASELINE_SECOND_STOP_STEP = "snort_severe_baseline_second_stop_step"
    SNORT_WARNING_BASELINE_SECOND_STOP_STEP = "snort_warning_baseline_second_stop_step"
    SNORT_CRITICAL_BASELINE_SECOND_STOP_STEP = "snort_critical_baseline_second_stop_step"
    VAR_LOG_BASELINE_SECOND_STOP_STEP = "var_log_baseline_second_stop_step"
    STEP_BASELINE_SECOND_STOP_STEP = "step_baseline_uncaught_second_stop_step"
    SNORT_SEVERE_BASELINE_THIRD_STOP_STEP = "snort_severe_baseline_third_stop_step"
    SNORT_WARNING_BASELINE_THIRD_STOP_STEP = "snort_warning_baseline_third_stop_step"
    SNORT_CRITICAL_BASELINE_THIRD_STOP_STEP = "snort_critical_baseline_third_stop_step"
    VAR_LOG_BASELINE_THIRD_STOP_STEP = "var_log_baseline_third_stop_step"
    STEP_BASELINE_THIRD_STOP_STEP = "step_baseline_uncaught_third_stop_step"
    SNORT_SEVERE_BASELINE_FOURTH_STOP_STEP = "snort_severe_baseline_fourth_stop_step"
    SNORT_WARNING_BASELINE_FOURTH_STOP_STEP = "snort_warning_baseline_fourth_stop_step"
    SNORT_CRITICAL_BASELINE_FOURTH_STOP_STEP = "snort_critical_baseline_fourth_stop_step"
    VAR_LOG_BASELINE_FOURTH_STOP_STEP = "var_log_baseline_fourth_stop_step"
    STEP_BASELINE_FOURTH_STOP_STEP = "step_baseline_uncaught_fourth_stop_step"
    SNORT_SEVERE_BASELINE_STOPS_REMAINING = "snort_severe_baseline_stops_remaining"
    SNORT_WARNING_BASELINE_STOPS_REMAINING = "snort_warning_baseline_stops_remaining"
    SNORT_CRITICAL_BASELINE_STOPS_REMAINING = "snort_critical_baseline_stops_remaining"
    VAR_LOG_BASELINE_STOPS_REMAINING = "var_log_baseline_stops_remaining"
    STEP_BASELINE_STOPS_REMAINING = "step_baseline_uncaught_stops_remaining"
    OPTIMAL_STOPS_REMAINING = "optimal_stops_remaining"
    OPTIMAL_FIRST_STOP_STEP = "optimal_first_stop_step"
    OPTIMAL_SECOND_STOP_STEP = "optimal_second_stop_step"
    OPTIMAL_THIRD_STOP_STEP = "optimal_third_stop_step"
    OPTIMAL_FOURTH_STOP_STEP = "optimal_fourth_stop_step"
    OPTIMAL_DEFENDER_EPISODE_STEPS = "optimal_defender_episode_steps"
    TERMINAL_OBSERVATION = "terminal_observation"


class DOCKER:
    """
    Constants related to Docker
    """
    CONTAINER_EXIT_STATUS = "exited"
    CONTAINER_CREATED_STATUS = "created"
    IMAGE_CREATED = "Created"
    IMAGE_OS = "Os"
    IMAGE_SIZE = "Size"
    IMAGE = "Image"
    IMAGE_ARCHITECTURE = "Architecture"
    REPO_TAGS = "RepoTags"
    ENTRYPOINT = "Entrypoint"
    BASE_CONTAINER_TYPE = "base"
    CONTAINER_CONFIG_DIR = "dir"
    EMULATION = "emulation"
    KAFKA_CONFIG = "kafka_config"
    CFG = "cfg"
    CONTAINER_CONFIG_CFG = "containers_cfg"
    CONTAINER_CONFIG_FLAGS_CFG = "flags_cfg"
    CONTAINER_CONFIG_TOPOLOGY_CFG = "topology_cfg"
    CONTAINER_CONFIG_USERS_CFG = "users_cfg"
    CONTAINER_CONFIG_VULNERABILITIES_CFG = "vulnerabilities_cfg"
    CONTAINER_CONFIG_TRAFFIC_CFG = "traffic_cfg"
    CONTAINER_CONFIG_CFG_PATH = "/containers.json"
    EMULATION_ENV_CFG_PATH = "/config.json"
    EMULATION_ENV_IMAGE = "/env.png"
    SIMULATION_ENV_IMAGE = "/env.png"
    KAFKA_CFG_PATH = "/kafka_config.json"
    CONTAINER_CONFIG_FLAGS_CFG_PATH = "/flags.json"
    CONTAINER_CONFIG_TOPOLOGY_CFG_PATH = "/topology.json"
    CONTAINER_CONFIG_USERS_CFG_PATH = "/users.json"
    CONTAINER_CONFIG_RESOURCES_CFG_PATH = "/resources.json"
    CONTAINER_CONFIG_VULNERABILITIES_CFG_PATH = "/vulnerabilities.json"
    CONTAINER_CONFIG_TRAFFIC_CFG_PATH = "/traffic.json"
    NET_ADMIN = "NET_ADMIN"
    UNIX_DOCKER_SOCK_URL = "unix://var/run/docker.sock"
    SSH_PREFIX = "ssh://kim@"
    CREATE_FLAGS_SCRIPT = "./create_flags.py"
    CREATE_TOPOLOGY_SCRIPT = "./create_topology.py"
    CREATE_VULN_SCRIPT = "./create_vuln.py"
    CREATE_USERS_SCRIPT = "./create_users.py"
    CREATE_TRAFFIC_GENERATORS_SCRIPT = "./create_traffic_generators.py"
    LIST_NETWORKS_CMD = "docker network ls"
    INSPECT_DOCKER_GWBRIDGE = "docker network inspect docker_gwbridge"
    NETWORK_CONNECT = "docker network connect"
    LIST_RUNNING_CONTAINERS_CMD = "docker ps -q"
    INSPECT_CONTAINER_CONFIG_CMD = "docker inspect"
    MAKEFILE = "Makefile"
    MAKEFILE_PATH = "/Makefile"
    NETWORK_SETTINGS = "NetworkSettings"
    NETWORKS = "Networks"
    CREATED_INFO = "Created"
    CONFIG = "Config"
    IMAGE_INFO = "Info"
    IP_ADDRESS_INFO = "IPAddress"
    NETWORK_ID_INFO = "NetworkID"
    GATEWAY_INFO = "Gateway"
    MAC_ADDRESS_INFO = "MacAddress"
    IP_PREFIX_LEN_INFO = "IPPrefixLen"
    HOSTNAME_INFO = "Hostname"
    CONTAINERS_DIR = "containers"
    CONTAINERS_KEY = "Containers"
    IPV4_KEY = "IPv4Address"
    CONTAINER_MAKEFILE_TEMPLATE_NAME = "Container_Makefile_template"
    CONTAINER_MAKEFILE_TEMPLATE_DIR_RELATIVE = "/../../../common/"
    MAKEFILE_TEMPLATE = "Makefile_template"
    ON_FAILURE_3 = "on-failure:3"
    CONTAINER_MAKEFILE_TEMPLATE_STR = \
        "\nall: run\n\nrun:\n\tdocker container run -dt " \
        "--name $(PROJECT)-$(CONTAINER)$(SUFFIX)-level$(LEVEL) " \
        "--hostname=$(CONTAINER)$(SUFFIX) --label dir=$(DIR) --label cfg=$(CFG) --label emulation=$(EMULATION)" \
        "--network=none --publish-all=true --memory=$(MEMORY) " \
        "-e TZ=Europe/Stockholm " \
        "--cpus=$(NUM_CPUS) --restart=$(RESTART_POLICY) " \
        "--cap-add NET_ADMIN $(PROJECT)/$(CONTAINER):$(VERSION)\n\nshell:\n\t" \
        "docker exec -it $(PROJECT)-$(CONTAINER)$(SUFFIX)-level$(LEVEL) /bin/bash\n\nstart:\n\t" \
        "docker container start $(PROJECT)-$(CONTAINER)$(SUFFIX)-level$(LEVEL)\n\nstop:\n\t" \
        "-docker stop $(PROJECT)-$(CONTAINER)$(SUFFIX)-level$(LEVEL)\n\nclean: stop\n\t" \
        "-docker rm $(PROJECT)-$(CONTAINER)$(SUFFIX)-level$(LEVEL)"
    BRIDGE_NETWORK_DRIVER = "bridge"
    OVERLAY_NETWORK_DRIVER = "overlay"
    ATTACHABLE_NETWORK_FLAG = "--attachable"


class CSLE:
    """
    Constants related to general CSLE
    """
    NAME = "csle"
    LEVEL = "level"
    BRIDGE = "br"
    CSLE_NETWORK_PREFIX = "csle_net_"
    CSLE_SUBNETMASK_PREFIX = "<EXECUTION_ID>."
    CSLE_FIRST_IP_OCTET_PLACEHOLDER = "<EXECUTION_ID>"
    CSLE_LEVEL_SUBNETMASK_SUFFIX = ".0.0/16"
    CSLE_EDGE_SUBNETMASK_SUFFIX = ".0/24"
    EDGE_SUBNETMASK_BITS = 24
    CSLE_EDGE_BITMASK = "255.255.255.0"
    CSLE_BITMASK = "255.255.0.0"
    NON_IDS_ROUTER = "router_1"
    LIST_OF_IP_SUBNETS = list(range(15, 171)) + list(range(173, 191)) + list(range(193, 250))


class MANAGEMENT:
    """
    Constants related to the management system
    """
    LIST_STOPPED = "list_stopped"
    LIST_RUNNING = "list_running"
    LIST_IMAGES = "list_images"
    STOP_RUNNING = "stop_running"
    RM_STOPPED = "rm_stopped"
    RM_IMAGES = "rm_images"
    RM_NETWORKS = "rm_networks"
    START_STOPPED = "start_stopped"
    LIST_NETWORKS = "list_networks"
    CLEAN = "clean"
    CLEAN_CONFIG = "clean_config"
    GEN_CONFIG = "gen_config"
    APPLY_CONFIG = "apply_config"
    RUN = "run"
    STOP = "stop"
    START = "start"
    TOPOLOGY = "topology"
    USERS = "users"
    FLAGS = "flags"
    VULN = "vuln"
    ALL = "all"
    CLEAN_FS_CACHE = "clean_fs_cache"
    TRAFFIC = "traffic"
    CLEAN_ENVS = "clean_envs"


class MAKEFILE:
    """
    Constants related to Makefiles
    """
    PROJECT = "PROJECT"
    INTERNAL_NETWORK = "INTERNAL_NETWORK"
    EXTERNAL_NETWORK = "EXTERNAL_NETWORK"
    EMULATION = "EMULATION"
    CONTAINER = "CONTAINER"
    VERSION = "VERSION"
    LEVEL = "LEVEL"
    DIR = "DIR"
    CFG = "CFG"
    FLAGSCFG = "FLAGSCFG"
    TOPOLOGYCFG = "TOPOLOGYCFG"
    USERSCFG = "USERSCFG"
    VULNERABILITIESCFG = "VULNERABILITIESCFG"
    SUFFIX = "SUFFIX"
    RESTART_POLICY = "RESTART_POLICY"
    NUM_CPUS = "NUM_CPUS"
    MEMORY = "MEMORY"


class VULNERABILITY_GENERATOR:
    """
    Constants related to the vulnerability generator
    """
    NAMES_SHORTLIST = ["admin", "test", "guest", "info", "adm", "mysql", "user", "administrator",
                       "oracle", "ftp", "pi", "puppet", "ansible", "ec2-user", "vagrant", "azureuser",
                       "donald", "alan"]


class EXPERIMENT:
    """
    Constants related to experiments folder structure
    """
    RESULTS_DIR = "results"
    DATA_DIR = "data"
    VIDEOS_DIR = "videos"
    GIFS_DIR = "gifs"
    TENSORBOARD_DIR = "tensorboard"
    ENV_DATA_DIR = "env_data"
    LOG_DIR = "logs"
    HYPERPARAMETERS_DIR = "hyperparameters"
    PLOTS_DIR = "plots"
    CONFIG_FILE_PATH = "/config.json"


class METADATA_STORE:
    """
    String constants related to the metadata store
    """
    DBNAME = "csle"
    USER = "csle"
    PASSWORD = "csle"
    HOST = "172.31.212.92"
    TRACES_PROPERTY = "traces"
    DB_NAME_PROPERTY = "dbname"
    PW_PROPERTY = "password"
    HOST_PROPERTY = "host"
    USER_PROPERTY = "user"
    EMULATIONS_TABLE = "emulations"
    SIMULATIONS_TABLE = "simulations"
    EMULATION_TRACES_TABLE = "emulation_traces"
    SIMULATION_TRACES_TABLE = "simulation_traces"
    EMULATION_STATISTICS_TABLE = "emulation_statistics"
    EMULATION_IMAGES_TABLE = "emulation_images"
    SIMULATION_IMAGES_TABLE = "simulation_images"
    EMULATION_SIMULATION_TRACES_TABLE = "emulation_simulation_traces"
    EXPERIMENT_EXECUTIONS_TABLE = "experiment_executions"
    TRAINING_JOBS_TABLE = "training_jobs"
    SYSTEM_IDENTIFICATION_JOBS_TABLE = "system_identification_jobs"
    DATA_COLLECTION_JOBS_TABLE = "data_collection_jobs"
    MULTI_THRESHOLD_STOPPING_POLICIES_TABLE = "multi_threshold_stopping_policies"
    LINEAR_THRESHOLD_STOPPING_POLICIES_TABLE = "linear_threshold_stopping_policies"
    PPO_POLICIES_TABLE = "ppo_policies"
    GAUSSIAN_MIXTURE_SYSTEM_MODELS_TABLE = "gaussian_mixture_system_models"
    TABULAR_POLICIES_TABLE = "tabular_policies"
    ALPHA_VEC_POLICIES_TABLE = "alpha_vec_policies"
    DQN_POLICIES_TABLE = "dqn_policies"
    FNN_W_SOFTMAX_POLICIES_TABLE = "fnn_w_softmax_policies"
    VECTOR_POLICIES_TABLE = "vector_policies"
    EMULATION_EXECUTIONS_TABLE = "emulation_executions"
    EMPIRICAL_SYSTEM_MODELS_TABLE = "empirical_system_models"
    MCMC_SYSTEM_MODELS_TABLE = "mcmc_system_models"
    GP_SYSTEM_MODELS_TABLE = "gp_system_models"
    CONFIG_TABLE = "config"
    MANAGEMENT_USERS_TABLE = "management_users"
    SESSION_TOKENS_TABLE = "session_tokens"
    TRACES_DATASETS_TABLE = "traces_datasets"
    STATISTICS_DATASETS_TABLE = "statistics_datasets"
    NUM_RETRIES_UPDATE_TRAINING_JOB = 5


class CONTAINER_POOLS:
    """
    Constants related to container pools
    """
    CONTAINER_POOL = [(f"{CONTAINER_IMAGES.FTP_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.FTP_2}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.HONEYPOT_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.HONEYPOT_2}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SSH_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SSH_2}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SSH_3}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.TELNET_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.TELNET_2}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.TELNET_3}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2015_1427_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2015_3306_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2016_10033_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SAMBA_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SQL_INJECTION_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SHELLSHOCK_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2010_0426_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2015_5602_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2014_0160_1}", "0.0.1")
                      ]

    GW_VULN_CONTAINERS = [(f"{CONTAINER_IMAGES.SSH_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SSH_2}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SSH_3}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.TELNET_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.TELNET_2}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.TELNET_3}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.CVE_2015_1427_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.CVE_2015_3306_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.CVE_2016_10033_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SAMBA_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SQL_INJECTION_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SHELLSHOCK_1}", "0.0.1")
                          ]

    PW_VULN_CONTAINERS = [(f"{CONTAINER_IMAGES.SSH_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SSH_2}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.SSH_3}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.TELNET_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.TELNET_2}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.TELNET_3}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.FTP_1}", "0.0.1"),
                          (f"{CONTAINER_IMAGES.FTP_2}", "0.0.1")
                          ]
    RCE_CONTAINERS = [(f"{CONTAINER_IMAGES.CVE_2015_1427_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2015_3306_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.CVE_2016_10033_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SAMBA_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SQL_INJECTION_1}", "0.0.1"),
                      (f"{CONTAINER_IMAGES.SHELLSHOCK_1}", "0.0.1")
                      ]
    SQL_INJECTION_CONTAINERS = [(f"{CONTAINER_IMAGES.SQL_INJECTION_1}", "0.0.1")]
    PRIV_ESC_CONTAINERS = [(f"{CONTAINER_IMAGES.CVE_2010_0426_1}", "0.0.1"),
                           (f"{CONTAINER_IMAGES.CVE_2015_5602_1}", "0.0.1")]

    AGENT_CONTAINERS = [((f"{CONTAINER_IMAGES.HACKER_KALI_1}", "0.0.1"))]
    ROUTER_CONTAINERS = [(f"{CONTAINER_IMAGES.ROUTER_1}", "0.0.1"),
                         (f"{CONTAINER_IMAGES.ROUTER_2}", "0.0.1")]


class ENV_CONSTANTS:
    """
    Constants  related to emulation environments
    """
    ATTACKER_SSH_RETRY_FIND_FLAG = 5
    ATTACKER_RETRY_FIND_USERS = 5
    ATTACKER_FTP_RETRY_FIND_FLAG = 2
    ATTACKER_RETRY_INSTALL_TOOLS = 5
    ATTACKER_INSTALL_TOOLS_SLEEP_SECONDS = 3
    ATTACKER_RETRY_CHECK_ROOT = 3
    ATTACKER_RETRY_SAMBACRY = 4
    ATTACKER_RETRY_SHELLSHOCK = 4
    ATTACKER_RETRY_DVWA_SQL_INJECTION = 10
    ATTACKER_RETRY_CVE_2015_3306 = 4
    ATTACKER_RETRY_CVE_2015_1427 = 4
    ATTACKER_RETRY_CVE_2016_10033 = 4
    ATTACKER_RETRY_CVE_2010_0426 = 4
    ATTACKER_RETRY_CVE_2015_5602 = 4
    ATTACKER_SAMBACRY_SLEEP_RETRY = 4
    ATTACKER_SHELLSHOCK_SLEEP_RETRY = 4
    ATTACKER_DVWA_SQL_INJECTION_SLEEP_RETRY = 4
    ATTACKER_CVE_2015_3306_SLEEP_RETRY = 4
    ATTACKER_CVE_2015_1427_SLEEP_RETRY = 4
    ATTACKER_CVE_2016_10033_SLEEP_RETRY = 4
    ATTACKER_CVE_2010_0426_SLEEP_RETRY = 4
    ATTACKER_CVE_2015_5602_SLEEP_RETRY = 4
    ATTACKER_CONTINUE_ACTION_SLEEP = 0.001
    SHELL_READ_WAIT = 0.5
    SHELL_MAX_TIMEOUTS = 4000
    MAX_NMAP_COMMAND_OUTPUT_SIZE = 10000000
    RETRY_TIMEOUT = 2
    NUM_RETRIES = 5
    SLEEP_RETRY = 5


class LOGGING:
    """
    Constants related to logging
    """
    DEFAULT_LOG_DIR = "/tmp/csle/"


class SIMULATION:
    """
    Constants related to simulations
    """
    SIMULATION_ENV_CFG_PATH = "/config.json"


class STATIC_ATTACKERS:
    """
    Constants related to static attackers
    """
    EXPERT = "expert"
    EXPERIENCED = "experienced"
    NOVICE = "novice"


class OPENFLOW:
    """
    Constants related to OPENFLOW
    """
    OPENFLOW_V_1_0 = "OpenFlow10"
    OPENFLOW_V_1_1 = "OpenFlow11"
    OPENFLOW_V_1_2 = "OpenFlow12"
    OPENFLOW_V_1_3 = "OpenFlow13"
    OPENFLOW_V_1_4 = "OpenFlow14"
    OPENFLOW_V_1_5 = "OpenFlow15"


class DATASETS:
    """
    Constants related to datasets
    """
    METADATA_FILE_NAME = "readme.json"
    JSON_FILE_FORMAT = "json"
    FILE_FORMAT_PROPERTY = "file_format"
    NUM_TRACES_PROPERTY = "num_traces"
    NUM_ATTRIBUTES_PER_TIME_STEP_PROPERTY = "num_attributes_per_time_step"
    SCHEMA_PROPERTY = "schema"
    NUM_TRACES_PER_FILE_PROPERTY = "num_traces_per_file"
    ADDED_BY_PROPERTY = "added_by"
    COLUMNS_PROPERTY = "columns"
    NUM_MEASUREMENTS_PROPERTY = "num_measurements"
    NUM_CONDITIONS_PROPERTY = "num_conditions"
    NUM_METRICS_PROPERTY = "num_metrics"
    CONDITIONS_PROPERTY = "conditions"
    METRICS_PROPERTY = "metrics"


class T_SPSA:
    """
    String constants related to T-SPSA
    """
    a = "a"
    c = "c"
    LAMBDA = "lambda"
    POLICY_TYPE = "policy_type"
    OBJECTIVE_TYPE = "objective_type"
    A = "A"
    EPSILON = "epsilon"
    N = "N"
    L = "L"
    IMPROVE_BREAK = "improve_break"
    THETA1 = "theta1"
    THETAS = "thetas"
    THRESHOLDS = "thresholds"
    GRADIENT_BATCH_SIZE = "gradient_batch_size"
    STOP_DISTRIBUTION_ATTACKER = "stop_distribution_attacker"
    STOP_DISTRIBUTION_DEFENDER = "stop_distribution_defender"


class NEURAL_NETWORKS:
    """
    Constants related to neural networks
    """
    NUM_NEURONS_PER_HIDDEN_LAYER = "num_neurons_per_hidden_layer"
    NUM_HIDDEN_LAYERS = "num_hidden_layers"
    ACTIVATION_FUNCTION = "activation_function"
    DEVICE = "device"


class CUDA:
    """
    String constants related to cuda
    """
    CUDA_VISIBLE_DEVICES = "CUDA_VISIBLE_DEVICES"



from unittest.mock import patch, MagicMock
from csle_common.util.experiment_util import ExperimentUtil
import csle_common.constants.constants as constants
import tempfile
import os
import shutil
import sys
import numpy as np
import random
import torch


class TestExperimentUtilSuite:
    """
    Test suite for experiment util
    """

    def test_get_subdir(self) -> None:
        """
        Test the function that constructs the subdir string from a given results dir, subdir, and random seed

        :return: None
        """
        output_dir = "output"
        results_dir = "results"
        subdir = "logs"
        seed = 10

        expected = "output/results/logs/10/"
        result = ExperimentUtil.get_subdir(output_dir=output_dir, results_dir=results_dir, subdir=subdir, seed=seed)

        assert expected == result

    def test_create_artifact_dirs(self) -> None:
        """
        Test the method that creates artifact directories if they do not already exist

        :return: None
        """
        temp_dir = tempfile.mkdtemp()
        try:
            random_seed = 10
            ExperimentUtil.create_artifact_dirs(output_dir=temp_dir, random_seed=random_seed)
            expected_dirs = [
                constants.EXPERIMENT.LOG_DIR,
                constants.EXPERIMENT.PLOTS_DIR,
                constants.EXPERIMENT.DATA_DIR,
                constants.EXPERIMENT.HYPERPARAMETERS_DIR,
                constants.EXPERIMENT.GIFS_DIR,
                constants.EXPERIMENT.TENSORBOARD_DIR,
                constants.EXPERIMENT.ENV_DATA_DIR,
            ]
            for subdir in expected_dirs:
                subdir_path = os.path.join(temp_dir, constants.EXPERIMENT.RESULTS_DIR, subdir, str(random_seed))
                assert os.path.exists(subdir_path), f"Directory {subdir_path} does not exist"

        finally:
            shutil.rmtree(temp_dir)

    def test_setup_experiment_logger(self, tmpdir) -> None:
        """
        Test the function that configures the logger for writing log-data of training

        :param tmpdir: temporary directory
        :return: None
        """
        logger_name = "test_logger"
        logdir = str(tmpdir.mkdir("logs"))
        time_str = "20240101-1234"

        logger = ExperimentUtil.setup_experiment_logger(name=logger_name, logdir=logdir, time_str=time_str)
        assert logger.name == logger_name

    @patch("io.open")
    def test_write_emulation_config_file(self, mock_open) -> None:
        """
        Test the function that writes a config object to a config file

        :param mock_open: mock_open
        :return: None
        """
        emulation_env_config = MagicMock()
        emulation_env_config.to_dict.return_value = {"key1": "value1", "key2": "value2"}
        ExperimentUtil.write_emulation_config_file(emulation_env_config=emulation_env_config, path="path")
        mock_open.assert_called()

    @patch("io.open")
    def test_write_simulation_config_file(self, mock_open) -> None:
        """
        Test the function that writes a config object to a config file

        :param mock_open: mock_open
        :return: None
        """
        simulation_env_config = MagicMock()
        simulation_env_config.to_dict.return_value = {"key1": "value1", "key2": "value2"}
        ExperimentUtil.write_simulation_config_file(simulation_env_config=simulation_env_config, path="path")
        mock_open.assert_called()

    @patch("builtins.open")
    def test_read_env_picture(self, mock_open) -> None:
        """
        Mock the method that reads the environment topology picture from a file

        :param mock_open: mock_open
        :return: None
        """
        mock_open.return_value.__enter__.return_value.read.return_value = b"fake_image_data"
        image_data = ExperimentUtil.read_env_picture("fake_path")
        assert image_data == b"fake_image_data"

    @patch.object(sys, "argv", ["file_name"])
    def test_parse_args(self) -> None:
        """
        Test the method that parses the commandline arguments with argparse

        :return: None
        """
        default_config_path = "default/path/to/config.json"
        args = ExperimentUtil.parse_args(default_config_path)
        assert args.configpath == default_config_path

    @patch.object(sys, "argv", ["/fake/path/to/script.py"])
    def test_get_script_path(self) -> None:
        """
        Test the method that returns the script path

        :return: None
        """
        expected_path = "/fake/path/to"
        script_path = ExperimentUtil.get_script_path()
        assert script_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.get_script_path")
    def test_default_output_dir(self, mock_get_script_path) -> None:
        """
        Test the method that returns the default output dir

        :param mock_get_script_path: mock_get_script_path
        :return: None
        """
        mock_get_script_path.return_value = "path"
        expected_path = "path"
        output_dir = ExperimentUtil.default_output_dir()
        assert output_dir == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def test_default_emulation_config_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to emulation config file

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.DOCKER.EMULATION_ENV_CFG_PATH = "fake_emulation_config.json"
        expected_path = "/fake/default/output/dir/.fake_emulation_config.json"
        config_path = ExperimentUtil.default_emulation_config_path()
        assert config_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def test_default_simulation_config_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to simulation config file

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.SIMULATION.SIMULATION_ENV_CFG_PATH = "fake_simulation_config.json"
        expected_path = "/fake/default/output/dir/.fake_simulation_config.json"
        config_path = ExperimentUtil.default_simulation_config_path()
        assert config_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def test_default_emulation_picture_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to emulation picture file

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.DOCKER.EMULATION_ENV_IMAGE = "fake_emulation_picture.img"
        expected_path = "/fake/default/output/dir/.fake_emulation_picture.img"
        picture_path = ExperimentUtil.default_emulation_picture_path()
        assert picture_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def test_default_simulation_picture_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to simulation picture file

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.DOCKER.SIMULATION_ENV_IMAGE = "fake_simulation_picture.img"
        expected_path = "/fake/default/output/dir/.fake_simulation_picture.img"
        picture_path = ExperimentUtil.default_simulation_picture_path()
        assert picture_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def test_default_containers_folders_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to container folders

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.COMMANDS.SLASH_DELIM = "/"
        constants.DOCKER.CONTAINERS_DIR = "container"
        expected_path = "/fake/default/output/dir/./container"
        container_path = ExperimentUtil.default_containers_folders_path()
        assert container_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def default_makefile_template_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to makefile tempalte

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.COMMANDS.SLASH_DELIM = "/"
        constants.DOCKER.MAKEFILE_TEMPLATE = "makefile"
        expected_path = "/fake/default/output/dir/./makefile"
        config_path = ExperimentUtil.default_makefile_template_path()
        assert config_path == expected_path

    @patch("csle_common.util.experiment_util.ExperimentUtil.default_output_dir")
    def default_makefile_path(self, mock_default_output_dir) -> None:
        """
        Test the method that returns the default path to makefile tempalte

        :param mock_default_output_dir: mock_default_output_dir
        :return: None
        """
        mock_default_output_dir.return_value = "/fake/default/output/dir"
        constants.COMMANDS.DOT_DELIM = "."
        constants.COMMANDS.SLASH_DELIM = "/"
        constants.DOCKER.MAKEFILE = "makefile"
        expected_path = "/fake/default/output/dir/./makefile"
        config_path = ExperimentUtil.default_makefile_path()
        assert config_path == expected_path

    def test_running_average_basic(self) -> None:
        """
        Test the method that used to compute the running average of the last N elements of a vector x

        :return: None
        """
        x = np.array([1, 2])
        N = 3
        result = ExperimentUtil.running_average(x, N)
        expected_result = 1
        assert result == expected_result

    def test_running_average_list(self) -> None:
        """
        Test the method that used to compute the running average of the last N elements of a vector x

        :return: None
        """
        x = [1, 2, 3, 4, 5]
        N = 3
        result = ExperimentUtil.running_average_list(x, N)
        expected_result = [1, 2, 2, 3, 3]
        assert np.allclose(result, expected_result)

    def test_mean_confidence_interval(self) -> None:
        """
        Test the method that compute confidence intervals

        :return: None
        """
        data = [1, 2, 3, 4, 5]
        m, _ = ExperimentUtil.mean_confidence_interval(data, 0.95)
        assert np.isclose(m, 3.0)

    def test_set_seed(self) -> None:
        """
        Test the method that sets the random seed

        :return: None
        """
        seed = 42

        ExperimentUtil.set_seed(seed)
        random_numbers1 = [random.random() for _ in range(5)]
        np_random_numbers1 = np.random.rand(5)
        torch_random_numbers1 = torch.rand(5)

        ExperimentUtil.set_seed(seed)
        random_numbers2 = [random.random() for _ in range(5)]
        np_random_numbers2 = np.random.rand(5)
        torch_random_numbers2 = torch.rand(5)

        assert random_numbers1 == random_numbers2
        assert np.allclose(np_random_numbers1, np_random_numbers2)
        assert torch.equal(torch_random_numbers1, torch_random_numbers2)

    def test_regress_lists(self) -> None:
        """
        Test the method that regress sublists.

        :return: None
        """
        lists = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
        result = ExperimentUtil.regress_lists(lists)
        expected_result = [[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]]
        assert np.allclose(result[0], expected_result[0])
        assert np.allclose(result[1], expected_result[1])
        assert np.allclose(result[2], expected_result[2])



from typing import Set, List, Dict, Any, Tuple
from csle_common.dao.emulation_config.default_network_firewall_config import DefaultNetworkFirewallConfig
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeFirewallConfig(JSONSerializable):
    """
    A DTO object representing a firewall configuration of a container in an emulation environment
    """

    def __init__(self, ips_gw_default_policy_networks: List[DefaultNetworkFirewallConfig],
                 hostname: str, output_accept: Set[str], input_accept: Set[str],
                 forward_accept: Set[str], output_drop: Set[str], input_drop: Set[str],
                 forward_drop: Set[str], routes: Set[Tuple[str, str]], docker_gw_bridge_ip: str = "",
                 physical_host_ip: str = ""):
        """
        Initializes the DTO

        :param ips_gw_default_policy_networks: List of ip,gw,default policy, network
        :param ip: the ip of the node
        :param hostname: the hostname of the node
        :param output_accept: the list of ips to accept output
        :param input_accept: the list of ips to accept input
        :param forward_accept: the list of ips to accept forward
        :param output_drop: the list of ips to drop output
        :param input_drop: the list of ips to drop input
        :param forward_drop: the list of ips to drop forward
        :param routes: the set of custom routes for the routing table
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.ips_gw_default_policy_networks = ips_gw_default_policy_networks
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.hostname = hostname
        self.output_accept = output_accept
        self.input_accept = input_accept
        self.forward_accept = forward_accept
        self.output_drop = output_drop
        self.input_drop = input_drop
        self.forward_drop = forward_drop
        self.routes = routes
        self.physical_host_ip = physical_host_ip

    def get_ips(self):
        """
        :return: list of ip addresses
        """
        return list(filter(lambda x: x is not None, map(lambda x: x.ip, self.ips_gw_default_policy_networks)))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeFirewallConfig":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = NodeFirewallConfig(
            hostname=d["hostname"],
            ips_gw_default_policy_networks=list(map(lambda x: DefaultNetworkFirewallConfig.from_dict(x),
                                                    d["ips_gw_default_policy_networks"])),
            output_accept=set(d["output_accept"]),
            input_accept=set(d["input_accept"]),
            forward_accept=set(d["forward_accept"]),
            output_drop=set(d["output_drop"]),
            input_drop=set(d["input_drop"]),
            forward_drop=set(d["forward_drop"]),
            routes=set(list(map(lambda x: (str(x[0]), str(x[1])), d["routes"]))),
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"],
            physical_host_ip=d["physical_host_ip"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["hostname"] = self.hostname
        d["ips_gw_default_policy_networks"] = list(map(lambda x: x.to_dict(), self.ips_gw_default_policy_networks))
        d["output_accept"] = list(self.output_accept)
        d["input_accept"] = list(self.input_accept)
        d["forward_accept"] = list(self.forward_accept)
        d["output_drop"] = list(self.output_drop)
        d["input_drop"] = list(self.input_drop)
        d["forward_drop"] = list(self.forward_drop)
        d["routes"] = list(self.routes)
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"ips_gw_default_policy_networks:{list(map(lambda x: str(x), self.ips_gw_default_policy_networks))}, " \
               f"output_accept:{self.output_accept}, " \
               f"input_accept:{self.input_accept}, forward_accept:{self.forward_accept}, " \
               f"output_drop:{self.output_drop}, " \
               f"input_drop:{self.input_drop}, forward_drop:{self.forward_drop}, " \
               f"routers:{self.routes}, hostname: {self.hostname}, docker_gw_bridge_ip:{self.docker_gw_bridge_ip}," \
               f" physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeFirewallConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeFirewallConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeFirewallConfig":
        """
        :return: a copy of the DTO
        """
        return NodeFirewallConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "NodeFirewallConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.output_accept = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip(
                ip=x, ip_first_octet=ip_first_octet), list(config.output_accept))))
        config.input_accept = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip(
                ip=x, ip_first_octet=ip_first_octet), list(config.input_accept))))
        config.forward_accept = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip(
                ip=x, ip_first_octet=ip_first_octet), list(config.forward_accept))))
        config.output_drop = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip(
                ip=x, ip_first_octet=ip_first_octet), list(config.output_drop))))
        config.input_drop = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip(
                ip=x, ip_first_octet=ip_first_octet), list(config.input_drop))))
        config.forward_drop = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip(
                ip=x, ip_first_octet=ip_first_octet), list(config.forward_drop))))
        config.routes = set(list(map(
            lambda x: GeneralUtil.replace_first_octet_of_ip_tuple(
                tuple_of_ips=x, ip_first_octet=ip_first_octet), list(config.routes))))
        config.ips_gw_default_policy_networks = list(map(lambda x: x.create_execution_config(
            ip_first_octet=ip_first_octet), config.ips_gw_default_policy_networks))
        return config

    @staticmethod
    def schema() -> "NodeFirewallConfig":
        """
        :return: get the schema of the DTO
        """
        dto = NodeFirewallConfig(ips_gw_default_policy_networks=[DefaultNetworkFirewallConfig.schema()], hostname="",
                                 output_accept=set(), input_accept=set(), forward_accept=set(), output_drop=set(),
                                 input_drop=set(), forward_drop=set(), routes=set())
        dto.output_accept.add("")
        dto.input_accept.add("")
        dto.forward_accept.add("")
        dto.output_drop.add("")
        dto.input_drop.add("")
        dto.forward_drop.add("")
        dto.routes.add(("", ""))
        return dto



"""
Type of attack action outcomes
"""
from enum import IntEnum


class EmulationAttackerActionOutcome(IntEnum):
    """
    Enum representing the different attack outcomes in the emulation.
    """
    SHELL_ACCESS = 0
    INFORMATION_GATHERING = 1
    LOGIN = 2
    FLAG = 3
    PIVOTING = 4
    PRIVILEGE_ESCALATION_ROOT = 5
    GAME_END = 6
    CONTINUE = 7



from unittest.mock import patch, MagicMock
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.vulnerability_type import VulnType
from csle_common.controllers.vulnerabilities_controller import VulnerabilitiesController


class TestVulnControllerSuite:
    """
    Test VulnerabilitiesController
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    def test_create_vulns(self, mock_execute_ssh_cmd, mock_connect_admin) -> None:
        """
        Test method that creates vulnerabilities in an emulation environment

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        logger = MagicMock()
        credential = MagicMock(username="testuser")
        vuln = MagicMock(
            physical_host_ip="192.168.1.1", docker_gw_bridge_ip="192.168.1.2", vuln_type=VulnType.PRIVILEGE_ESCALATION,
            cve=constants.EXPLOIT_VULNERABILITES.CVE_2010_0426, credentials=[credential])
        emulation_env_config = MagicMock()
        emulation_env_config.vuln_config.node_vulnerability_configs = [vuln]
        emulation_env_config.connections = {"192.168.1.2": MagicMock()}
        physical_server_ip = "192.168.1.1"
        mock_execute_ssh_cmd.side_effect = [
            ("", "", 0),  # output of cp /etc/sudoers.bak /etc/sudoers
            ("", "", 0),  # output of adding CVE entry to sudoers
            ("", "", 0),  # output of chmod 440 /etc/sudoers
        ]
        VulnerabilitiesController.create_vulns(emulation_env_config, physical_server_ip, logger)
        mock_connect_admin.assert_any_call(emulation_env_config=emulation_env_config, ip="192.168.1.2")
        mock_execute_ssh_cmd.assert_called()



from typing import List, Dict, Any, Union
from csle_common.dao.emulation_config.node_beats_config import NodeBeatsConfig
from csle_base.json_serializable import JSONSerializable


class BeatsConfig(JSONSerializable):
    """
    A DTO object representing the beats configuration of an emulation environment
    """

    def __init__(self, node_beats_configs: List[NodeBeatsConfig], num_elastic_shards: int, reload_enabled: bool):
        """
        Initializes the DTO

        :param node_beats_configs: the list of node beats configurations
        :param num_elastic_shards: shards of elasticsearch
        :param reload_enabled: whether reload of beats configurations is enabled
        """
        self.node_beats_configs = node_beats_configs
        self.num_elastic_shards = num_elastic_shards
        self.reload_enabled = reload_enabled

    def get_node_beats_config_by_ips(self, ips: List[str]) -> Union[NodeBeatsConfig, None]:
        """
        Gets a node beats config which matches a list of ips

        :param ips: the ips
        :return: the node beats config or None
        """
        for node_beats_config in self.node_beats_configs:
            if node_beats_config.ip in ips:
                return node_beats_config
        return None

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "BeatsConfig":
        """
        Converts a dict representation of the object into a an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = BeatsConfig(
            node_beats_configs=list(map(lambda x: NodeBeatsConfig.from_dict(x), d["node_beats_configs"])),
            num_elastic_shards=d["num_elastic_shards"], reload_enabled=d["reload_enabled"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["node_beats_configs"] = list(map(lambda x: x.to_dict(), self.node_beats_configs))
        d["num_elastic_shards"] = self.num_elastic_shards
        d["reload_enabled"] = self.reload_enabled
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"node_beats_configs:{','.join(list(map(lambda x: str(x), self.node_beats_configs)))}, " \
               f"num_elastic_shards: {self.num_elastic_shards}, reload_enabled: {self.reload_enabled}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "BeatsConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return BeatsConfig.from_dict(json.loads(json_str))

    def copy(self) -> "BeatsConfig":
        """
        :return: a copy of the DTO
        """
        return BeatsConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int) -> "BeatsConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :return: the new config
        """
        config = self.copy()
        config.node_beats_configs = list(map(lambda x: x.create_execution_config(ip_first_octet=ip_first_octet),
                                             config.node_beats_configs))
        return config



from typing import List, Dict, Tuple, Union, Optional, Any
import random
import numpy as np
import csle_common.constants.constants as constants
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.simulation_config.state_type import StateType
from csle_common.dao.training.policy_type import PolicyType
from csle_common.dao.training.multi_threshold_stopping_policy import MultiThresholdStoppingPolicy


class MixedMultiThresholdStoppingPolicy(Policy):
    """
    A mixed multi-threshold stopping policy
    """

    def __init__(self, defender_Theta: List[List[List[float]]], attacker_Theta: List[List[List[List[float]]]],
                 simulation_name: str, L: int, states: List[State], player_type: PlayerType,
                 actions: List[Action], experiment_config: Optional[ExperimentConfig], avg_R: float,
                 agent_type: AgentType, opponent_strategy: Optional["MixedMultiThresholdStoppingPolicy"] = None):
        """
        Initializes the policy

        :param defender_Theta: the buffer with threshold vectors for the defender
        :param attacker_Theta: the buffer with threshold vectors for the attacker
        :param simulation_name: the simulation name
        :param attacker: whether it is an attacker or not
        :param L: the number of stop actions
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration used for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        :param agent_type: the agent type
        :param opponent_strategy: the opponent's strategy
        """
        super(MixedMultiThresholdStoppingPolicy, self).__init__(agent_type=agent_type, player_type=player_type)
        self.attacker_Theta = attacker_Theta
        self.defender_Theta = defender_Theta
        self.id = -1
        self.simulation_name = simulation_name
        self.L = L
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.opponent_strategy = opponent_strategy
        self.policy_type = PolicyType.MIXED_MULTI_THRESHOLD

    def probability(self, o: List[float], a: int) -> int:
        """
        Probability of a given action

        :param o: the current observation
        :param a: a given action
        :return: the probability of a
        """
        return self.action(o=o) == a

    def action(self, o: List[float], deterministic: bool = True) -> int:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        if not self.player_type == PlayerType.ATTACKER:
            a, _ = self._defender_action(o=o)
            return a
        else:
            if self.opponent_strategy is None:
                raise ValueError("Must specify the opponent strategy")
            a1, defender_stop_probability = self.opponent_strategy._defender_action(o=o)
            if a1 == 0:
                defender_stop_probability = 1 - defender_stop_probability
            a, _ = self._attacker_action(o=o, defender_stop_probability=defender_stop_probability)
            return a

    def _attacker_action(self, o: List[float], defender_stop_probability: float) -> Tuple[int, float]:
        """
        Multi-threshold stopping policy of the attacker

        :param o: the input observation
        :param defender_stop_probability: the defender's stopping probability
        :return: the selected action (int) and the probability of selecting that action
        """
        s = int(o[2])
        if s == 2:
            return 0, 1.0
        l = int(o[0])
        thresholds = self.attacker_Theta[s][l - 1][0]
        counts = self.attacker_Theta[s][l - 1][1]

        mixture_weights = np.array(counts) / sum(self.attacker_Theta[s][l - 1][1])
        prob = 0
        for i in range(len(thresholds)):
            if defender_stop_probability >= thresholds[i]:
                prob += mixture_weights[i] * MultiThresholdStoppingPolicy.stopping_probability(
                    b1=defender_stop_probability, threshold=thresholds[i], k=-20)

        if s == 1:
            a = 0
            if random.uniform(0, 1) < prob:
                a = 1
            else:
                prob = 1 - prob
        else:
            a = 1
            if random.uniform(0, 1) < prob:
                a = 0
            else:
                prob = 1 - prob
        prob = round(prob, 3)
        return a, prob

    def _defender_action(self, o) -> Tuple[int, float]:
        """
        Multi-threshold stopping policy of the defender

        :param o: the input observation
        :return: the selected action (int), and the probability of selecting that action
        """
        b1 = o[1]
        l = int(o[0])
        thresholds = self.defender_Theta[l - 1][0]
        counts = self.defender_Theta[l - 1][1]
        mixture_weights = np.array(counts) / sum(self.defender_Theta[l - 1][1])
        stop_probability = 0
        for i, thresh in enumerate(thresholds):
            if b1 >= thresh:
                stop_probability += mixture_weights[i] * MultiThresholdStoppingPolicy.stopping_probability(
                    b1=b1, threshold=thresh, k=-20)
        stop_probability = round(stop_probability, 3)

        if random.uniform(0, 1) >= stop_probability:
            return 0, 1 - stop_probability
        else:
            return 1, stop_probability

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["attacker_Theta"] = self.attacker_Theta
        d["defender_Theta"] = self.attacker_Theta
        if self.player_type == PlayerType.ATTACKER:
            d["Theta"] = self.attacker_Theta
        if self.player_type == PlayerType.DEFENDER:
            d["Theta"] = self.defender_Theta
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["player_type"] = self.player_type
        d["agent_type"] = self.agent_type
        d["L"] = self.L
        if self.experiment_config is not None:
            d["experiment_config"] = self.experiment_config.to_dict()
        else:
            d["experiment_config"] = None
        d["avg_R"] = self.avg_R
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "MixedMultiThresholdStoppingPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = MixedMultiThresholdStoppingPolicy(
            defender_Theta=d["defender_Theta"], attacker_Theta=d["attacker_Theta"],
            simulation_name=d["simulation_name"], L=d["L"],
            states=list(map(lambda x: State.from_dict(x), d["states"])), player_type=d["player_type"],
            actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
            experiment_config=ExperimentConfig.from_dict(d["experiment_config"]), avg_R=d["avg_R"],
            agent_type=d["agent_type"])
        obj.id = d["id"]
        return obj

    def _update_Theta_attacker(self, new_thresholds: List[List[List[float]]]) -> None:
        """
        Updates the Theta buffer with new attacker thresholds

        :param new_thresholds: the new thresholds to add to the buffer
        :return: None
        """
        for a_thresholds in new_thresholds:
            for s in range(2):
                for l in range(self.L):
                    if len(self.attacker_Theta[s][l]) == 0:
                        self.attacker_Theta[s][l] = [[a_thresholds[s][l]], [1]]
                    else:
                        if a_thresholds[s][l] in self.attacker_Theta[s][l][0]:
                            i = self.attacker_Theta[s][l][0].index(a_thresholds[s][l])
                            self.attacker_Theta[s][l][1][i] = self.attacker_Theta[s][l][1][i] + 1
                        else:
                            self.attacker_Theta[s][l][0].append(a_thresholds[s][l])
                            self.attacker_Theta[s][l][1].append(1)

    def _update_Theta_defender(self, new_thresholds: List[List[float]]) -> None:
        """
        Updates the theta buffer with new defender thresholds

        :param new_thresholds: the new thresholds to add to the buffer
        :return: None
        """
        for defender_theta in new_thresholds:
            for l in range(self.L):
                if len(self.defender_Theta[l]) == 0:
                    self.defender_Theta[l] = [[defender_theta[l]], [1]]
                else:
                    if defender_theta[l] in self.defender_Theta[l][0]:
                        i = self.defender_Theta[l][0].index(defender_theta[l])
                        self.defender_Theta[l][1][i] = self.defender_Theta[l][1][i] + 1
                    else:
                        self.defender_Theta[l][0].append(defender_theta[l])
                        self.defender_Theta[l][1].append(1)

    def stage_policy(self, o: List[Union[int, float]]) -> List[List[float]]:
        """
        Returns the stage policy for a given observation

        :param o: the observation to return the stage policy for
        :return: the stage policy
        """
        b1 = o[1]
        l = int(o[0])
        if not self.player_type == PlayerType.ATTACKER:
            stage_policy = []
            for _ in self.states:
                a1, stopping_probability = self._defender_action(o=o)
                if a1 == 0:
                    stopping_probability = 1 - stopping_probability
                stage_policy.append([1 - stopping_probability, stopping_probability])
            return stage_policy
        else:
            stage_policy = []
            if self.opponent_strategy is None:
                raise ValueError("Opponent strategy is None and must be specified.")
            a1, defender_stop_probability = self.opponent_strategy._defender_action(o=o)
            if a1 == 0:
                defender_stop_probability = 1 - defender_stop_probability
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    o = [l, b1, s.id]
                    a, action_prob = self._attacker_action(o=o, defender_stop_probability=defender_stop_probability)
                    if action_prob == 1:
                        action_prob = 0.99
                    if action_prob == 0:
                        action_prob = 0.01
                    if a == 1:
                        stage_policy.append([1 - action_prob, action_prob])
                    elif a == 0:
                        stage_policy.append([action_prob, 1 - action_prob])
                    else:
                        raise ValueError(f"Invalid state: {s.id}, valid states are: 0 and 1")
                else:
                    stage_policy.append([0.5, 0.5])
            return stage_policy

    def stop_distributions(self) -> Dict[str, List[float]]:
        """
        :return: the stop distributions and their names
        """
        distributions: Dict[str, List[float]] = {}
        if self.player_type == PlayerType.DEFENDER:
            belief_space = np.linspace(0, 1, num=100)
            for l in range(1, self.L + 1):
                stop_dist: List[float] = []
                for b in belief_space:
                    a1, prob = self._defender_action(o=[l, b])
                    if a1 == 1:
                        stop_dist.append(round(prob, 3))
                    else:
                        stop_dist.append(round(1 - prob, 3))
                distributions[constants.T_SPSA.STOP_DISTRIBUTION_DEFENDER + f"_l={l}"] = stop_dist
        else:
            defender_stop_space = np.linspace(0, 1, num=100)
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    for l in range(1, self.L + 1):
                        stop_dist = []
                        for pi_1_S in defender_stop_space:
                            a2, prob = self._attacker_action(o=[l, pi_1_S, s.id], defender_stop_probability=pi_1_S)
                            if a2 == 1:
                                stop_dist.append(round(prob, 3))
                            else:
                                stop_dist.append(round(1 - prob, 3))
                            distributions[constants.T_SPSA.STOP_DISTRIBUTION_ATTACKER + f"_l={l}_s={s.id}"] = stop_dist
        return distributions

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"defender_Theta: {self.defender_Theta}, attacker_theta: {self.attacker_Theta}" \
               f"id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"player_type: {self.player_type}, " \
               f"L:{self.L}, states: {self.states}, agent_type: {self.agent_type}, actions: {self.actions}," \
               f"experiment_config: {self.experiment_config}, avg_R: {self.avg_R}, policy_type: {self.policy_type}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "MixedMultiThresholdStoppingPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return MixedMultiThresholdStoppingPolicy.from_dict(json.loads(json_str))

    def copy(self) -> "MixedMultiThresholdStoppingPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())

    def __eq__(self, other: object) -> bool:
        """
        Compares equality with another object

        :param other: the object to compare with
        :return: True if equals, False otherwise
        """
        if not isinstance(other, MixedMultiThresholdStoppingPolicy):
            return False
        else:
            d1 = self.to_dict()
            d2 = other.to_dict()
            del d1["Theta"]
            del d2["Theta"]
            return d1 == d2



import select
import time
try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer
import csle_common.constants.constants as constants
from csle_common.logging.log import Logger


class ForwardSSHHandler(SocketServer.BaseRequestHandler):
    """
    SSH Server for forwarding local port over a SSH tunnel
    """

    def handle(self) -> None:
        """
        Main loop for handing the SSH connection

        :return: None
        """
        try:
            chan = self.server.ssh_transport.open_channel(
                constants.SSH.DIRECT_CHANNEL,
                (self.server.chain_host, self.server.chain_port),
                self.request.getpeername(),
            )
            while True:
                r, w, x = select.select([self.request, chan], [], [])
                if self.request in r:
                    try:
                        data = self.request.recv(1024)
                    except Exception as e:
                        if "Connection reset by peer" not in str(e):
                            Logger.__call__().get_logger().warning(f"forward SSH exception: {str(e)}, {repr(e)}."
                                                                   f"\nClosing the SSH tunnel.")
                            self.cleanup()
                        data = []
                    if len(data) == 0:
                        break
                    chan.send(data)
                if chan in r:
                    data = chan.recv(1024)
                    if len(data) == 0:
                        break
                    self.request.send(data)

            chan.close()
            self.request.close()
        except Exception as e:
            if "Transport endpoint" not in str(e):
                Logger.__call__().get_logger().warning(f"forward SSH exception2, {str(e)}, {repr(e)}."
                                                       f"\nClosing the SSH tunnel. ")
                self.cleanup()

    def cleanup(self) -> None:
        """
        Utility method for cleaning up the SSH tunnel

        :return: None
        """
        if self.server.chain_host in self.server.tunnels_dict:
            del self.server.tunnels_dict[self.server.chain_host]
        self.server.shutdown()
        time.sleep(0.5)  # wait for server to shutdown



from typing import Optional, List, Dict, Any
import gymnasium as gym
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_action.defender.emulation_defender_action_id import EmulationDefenderActionId
from csle_common.dao.emulation_action.defender.emulation_defender_stopping_actions \
    import EmulationDefenderStoppingActions
from csle_base.json_serializable import JSONSerializable


class EmulationDefenderActionConfig(JSONSerializable):
    """
    Configuration of the action space for the defender
    """

    def __init__(self, num_indices: int, actions: Optional[List[EmulationDefenderAction]] = None,
                 stopping_action_ids: Optional[List[EmulationDefenderActionId]] = None,
                 multiple_stop_actions: Optional[List[EmulationDefenderAction]] = None,
                 multiple_stop_actions_ids: Optional[List[EmulationDefenderActionId]] = None):
        """
        Class constructor

        :param num_indices: max num machine indexes allowed
        :param actions: list of actions in the action space
        :param stopping_action_ids: list of ids of the actions that are stopping actions
        :param multiple_stop_actions: if it is a multiple stopping environment, this defines the list of stop actions
        :param multiple_stop_actions_ids: if it is a multiple stopping environment, this defines the ids of
                                          the stop actions
        """
        self.actions = actions
        if self.actions is None:
            raise ValueError("Thera are no actions")
        self.num_actions = len(self.actions)
        self.num_indices = num_indices
        self.action_space = gym.spaces.Discrete(self.num_actions)
        self.action_lookup_d = {}
        self.action_lookup_d_val = {}
        if actions is not None:
            for action in actions:
                self.action_lookup_d[(action.id, action.index)] = action
                self.action_lookup_d_val[(action.id, action.index)] = action
        else:
            raise ValueError("actions is None and thus not iterable")

        self.stopping_action_ids = stopping_action_ids
        self.action_ids = self.stopping_action_ids
        self.multiple_stop_actions = multiple_stop_actions
        self.multiple_stop_actions_ids = multiple_stop_actions_ids
        if self.action_ids is not None:
            self.num_node_specific_actions = len(self.action_ids)
        else:
            raise ValueError("There are no number of node-specific actions")

    def print_actions(self) -> None:
        """
        Utility function for printing the list of actions

        :return: None
        """
        print("Defender Actions:")
        if self.actions is None:
            raise ValueError("self.actions is not iterable")
        else:
            for i, action in enumerate(self.actions):
                tag = "-"
                if not action.index == -1 and action.index is not None:
                    tag = str(action.index)
                else:
                    tag = "*"
                print(str(i) + ":" + action.name + "[" + tag + "]")

    def get_continue_action_idx(self) -> int:
        """
        :return: the index of the continue action
        """
        if self.actions is None:
            raise ValueError("self.actions is None and thus has no length")
        for i in range(len(self.actions)):
            if self.actions[i].id == EmulationDefenderActionId.CONTINUE:
                return i
        raise ValueError("No Continue Action in the action space")

    @staticmethod
    def all_actions_config(num_nodes: int, subnet_masks: List[str]) -> "EmulationDefenderActionConfig":
        """
        Creates an action configuration for the defender with all actions

        :param num_nodes: the number of nodes in the environment
        :param subnet_masks: the the subnet masks in the environment
        :return: the action configuration
        """
        defender_actions = []

        # Host actions
        for idx in range(num_nodes):
            pass

        # Subnet actions
        defender_actions.append(EmulationDefenderStoppingActions.STOP(index=num_nodes + 1))
        defender_actions.append(EmulationDefenderStoppingActions.CONTINUE(index=num_nodes + 1))

        defender_actions = sorted(defender_actions, key=lambda x: (x.id, x.index))
        stopping_action_ids = [
            EmulationDefenderActionId.STOP, EmulationDefenderActionId.CONTINUE
        ]
        defender_action_config = EmulationDefenderActionConfig(
            num_indices=num_nodes + 1, actions=defender_actions, stopping_action_ids=stopping_action_ids)
        return defender_action_config

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"num_indices: {self.num_indices}, actions: {self.actions}, " \
               f"stopping_action_ids: {self.stopping_action_ids}, " \
               f"multiple_stop_actions: {self.multiple_stop_actions}, " \
               f"multiple_stop_actions_ids: {self.multiple_stop_actions_ids}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation
        
        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["num_indices"] = self.num_indices
        d["actions"] = self.actions
        d["stopping_action_ids"] = self.stopping_action_ids
        d["multiple_stop_actions"] = self.multiple_stop_actions
        d["multiple_stop_actions_ids"] = self.multiple_stop_actions_ids
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationDefenderActionConfig":
        """
        Converts a dict representation to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationDefenderActionConfig(
            num_indices=d["num_indices"], actions=d["actions"],
            stopping_action_ids=d["stopping_action_ids"],
            multiple_stop_actions=d["multiple_stop_actions"],
            multiple_stop_actions_ids=d["multiple_stop_actions_ids"]
        )
        return obj

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationDefenderActionConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return EmulationDefenderActionConfig.from_dict(json.loads(json_str))



import os
from csle_common.logging.log import Logger
from csle_common.metastore.metastore_facade import MetastoreFacade
from csle_common.dao.system_identification.emulation_statistics import EmulationStatistics
from csle_common.dao.emulation_config.emulation_trace import EmulationTrace
from typing import Optional, Union


class ImportUtil:
    """
    Class with utility functions for importing data to the metastore
    """

    @staticmethod
    def import_emulation_statistics_from_disk_json(input_file: str, emulation_name: Union[None, str] = None) -> None:
        """
        Imports emulation statistics from disk to the metastore

        :param input file: the input file
        :param emulation_name: the emulation_name (optional)
        :return: None
        """
        Logger.__call__().get_logger().info(f"Importing emulation statistics from disk (json), input dir: {input_file}")
        if not os.path.exists(input_file):
            raise ValueError(f"File: {input_file} does not exist")
        statistics = EmulationStatistics.from_json_file(input_file)
        if emulation_name is not None:
            statistics.emulation_name = emulation_name
        MetastoreFacade.save_emulation_statistic(statistics)
        Logger.__call__().get_logger().info("Import of emulation statistics from disk complete, "
                                            f"input file:{input_file}")

    @staticmethod
    def import_emulation_traces_from_disk_json(input_file: str, emulation_name: Optional[str] = None) -> None:
        """
        Imports emulation traces from disk to the metastore

        :param input file: the input file
        :param emulation_name: the emulation_name (optional)
        :return: None
        """
        Logger.__call__().get_logger().info(f"Importing emulation traces from disk (json), input dir: {input_file}")
        if not os.path.exists(input_file):
            raise ValueError(f"File: {input_file} does not exist")
        traces = EmulationTrace.load_traces_from_disk(input_file)
        Logger.__call__().get_logger().info(f"Read {len(traces)} traces")
        for i, trace in enumerate(traces):
            Logger.__call__().get_logger().info(f"Saving trace {i}/{len(traces)} to the metastore")
            if emulation_name is not None:
                trace.emulation_name = emulation_name
            try:
                MetastoreFacade.save_emulation_trace(trace)
            except Exception as e:
                Logger.__call__().get_logger().info(f"There was an error saving trace number {i}, "
                                                    f"stacktrace: {e}, {repr(e)}")
        Logger.__call__().get_logger().info("Import of emulation traces from disk complete, "
                                            f"input file: {input_file}")
